import threading
from typing import Optional

from .airflow import AirflowRestClient
from .airflow.composer import ComposerRestClient
from .airflow.mwaa import MWAARestClient
from .databricks import DatabricksClient
from .s3 import S3Client
from ..utils.project_config import FabricInfo, FabricConfig


class RestClientFactory:
    _instance = None

    _s3_client_creation_lock = threading.Lock()

    def __init__(self, fabric_config: FabricConfig):

        self.fabric_config = fabric_config
        self.fabric_id_to_rest_client = {}

    def _get_fabric_info(self, fabric_id: str) -> FabricInfo:
        if fabric_id is None or fabric_id == "":
            raise ValueError(f"Fabric Id {fabric_id} is not defined in the deployment state")
        else:
            fabric_info: Optional[FabricInfo] = self.fabric_config.get_fabric(str(fabric_id))
            if fabric_info is None:
                raise ValueError(f"Fabric Id {fabric_id} is not found in the fabric configs {self.fabric_config}")
            return fabric_info

    def _get_client(self, fabric_id: str):
        if fabric_id in self.fabric_id_to_rest_client:
            return self.fabric_id_to_rest_client[fabric_id]
        else:
            return None

    def s3_client(self, fabric_id: str) -> S3Client:
        self._s3_client_creation_lock.acquire()
        try:

            if self._get_client(fabric_id) is not None:
                return self._get_client(fabric_id)

            emr = self._get_fabric_info(fabric_id).emr

            if emr is not None:
                client = S3Client(emr.region, emr.access_key_id, emr.secret_access_key, emr.session_token)
                self.fabric_id_to_rest_client[fabric_id] = client
                return client

            else:
                raise ValueError("Fabric Id is not defined in the deployment state")
        finally:
            self._s3_client_creation_lock.release()

    def databricks_client(self, fabric_id: str) -> DatabricksClient:
        if self._get_client(fabric_id) is not None:
            return self._get_client(fabric_id)

        databricks = self._get_fabric_info(fabric_id).databricks

        if databricks is not None:
            client = DatabricksClient.from_host_and_token(databricks.url, databricks.token)
            self.fabric_id_to_rest_client[fabric_id] = client
            return client

        else:
            raise ValueError(f"Fabric Id {fabric_id} is not defined and the databricks client {self._get_fabric_info(fabric_id)}.")

    def airflow_client(self, fabric_id: str) -> AirflowRestClient:
        if self._get_client(fabric_id) is not None:
            return self._get_client(fabric_id)

        fabric_info = self._get_fabric_info(fabric_id)

        composer = fabric_info.composer
        mwaa = fabric_info.mwaa

        if composer is not None:
            client = ComposerRestClient(composer.airflow_url, composer.project_id, composer.client_id,
                                        composer.key_json, composer.dag_location)
        elif mwaa is not None:
            client = MWAARestClient(mwaa.environment_name, mwaa.region, mwaa.access_key,
                                    mwaa.secret_key)

        else:
            raise ValueError(f"Fabric Id {fabric_id} is not defined.")

        if client is not None:
            self.fabric_id_to_rest_client[fabric_id] = client
            return client

    def dataproc_client(self, fabric_id: str) -> ComposerRestClient:
        if self._get_client(fabric_id) is not None:
            return self._get_client(fabric_id)

        fabric_info = self._get_fabric_info(fabric_id)

        dataproc = fabric_info.dataproc

        if dataproc is not None:
            client = ComposerRestClient("dummyAirflowUrl",
                                        dataproc.project_id,
                                        None,
                                        dataproc.key_json,
                                        "dummyDagLocation",
                                        dataproc.location)
        else:
            raise ValueError("Fabric Id is not defined in the deployment state")

        if client is not None:
            self.fabric_id_to_rest_client[fabric_id] = client
            return client

    @staticmethod
    def get_instance(cls, fabric_config):
        if cls._instance is None:
            cls._instance = RestClientFactory(fabric_config)

        return cls._instance



import tempfile
from typing import Dict, Optional

from databricks_cli.configure.provider import EnvironmentVariableConfigProvider
from databricks_cli.dbfs.api import DbfsApi
from databricks_cli.dbfs.dbfs_path import DbfsPath
from databricks_cli.jobs.api import JobsApi
from databricks_cli.sdk import ApiClient
from databricks_cli.sdk import JobsService
from databricks_cli.secrets.api import SecretApi
from requests import HTTPError
from tenacity import retry_if_exception_type, retry, stop_after_attempt, wait_exponential

from ..utility import Either
from ..utils.exceptions import DuplicateJobNameException


class DatabricksClient:
    def __init__(self, host: str = None, token: str = None):
        self.host = host
        self.token = token

        self.dbfs = DbfsApi(ApiClient(host=host, token=token, api_version="2.0"))
        self.job = JobsApi(ApiClient(host=host, token=token, api_version="2.1"))
        self.job_client = JobsService(ApiClient(host=host, token=token, api_version="2.1"))
        self.secret = SecretApi(ApiClient(host=host, token=token, api_version="2.0"))
        self.permission = PermissionsApi(ApiClient(host=host, token=token, api_version="2.0"))

    @classmethod
    def from_environment_variables(cls):
        config = EnvironmentVariableConfigProvider()
        host = config.get_config().host
        token = config.get_config().token
        return cls(host, token)

    @classmethod
    def from_host_and_token(cls, host: str, token: str):
        return cls(host, token)

    def upload_content(self, content: str, path: str):
        with tempfile.NamedTemporaryFile() as temp_file:
            temp_file.write(content.encode())
            # copy-pasted from gpt.
            temp_file.flush()  # Ensure that the content is flushed to disk
            temp_file.seek(0)  # Reset file's pointer to the beginning
            #
            # print(temp_file.read().decode())

            self.upload_src_path(src_path=temp_file.name, destination_path=path)

    @retry(retry=retry_if_exception_type(HTTPError), stop=stop_after_attempt(5),
           wait=wait_exponential(multiplier=2, max=30), reraise=True)
    def upload_src_path(self, src_path: str, destination_path: str):
        self.dbfs.put_file(src_path=src_path, dbfs_path=DbfsPath(destination_path, False), overwrite=True)

    def path_exist(self, path: str) -> bool:
        return self.dbfs.get_status(DbfsPath(path)) is not None

    def delete(self, path: str):
        self.dbfs.delete(path, recursive=True)

    @retry(retry=retry_if_exception_type(HTTPError), stop=stop_after_attempt(5),
           wait=wait_exponential(multiplier=2, max=30), reraise=True)
    def create_secret_scope_if_not_exist(self, secret_scope: str):

        response = self.secret.list_scopes()

        if any(scope['name'] == secret_scope for scope in response['scopes']) is False:
            self.secret.create_scope(secret_scope, initial_manage_principal="users", scope_backend_type=None,
                                     backend_azure_keyvault=None)
        else:
            return True

    def create_secret(self, scope: str, key: str, value: str):
        self.secret.put_secret(scope, key, value, None)

    @retry(retry=retry_if_exception_type(HTTPError), stop=stop_after_attempt(5),
           wait=wait_exponential(multiplier=2, max=30), reraise=True)
    def create_job(self, content: Dict[str, str]):
        return self.job.create_job(content)

    @retry(retry=retry_if_exception_type(HTTPError), stop=stop_after_attempt(5),
           wait=wait_exponential(multiplier=2, max=30), reraise=True)
    def get_job(self, scheduler_job_id: str):
        return self.job.get_job(scheduler_job_id)

    def delete_job(self, job_id: str):
        self.job.delete_job(job_id)

    @retry(retry=retry_if_exception_type(HTTPError), stop=stop_after_attempt(5),
           wait=wait_exponential(multiplier=2, max=30), reraise=True)
    def pause_job(self, scheduler_job_id: str) -> Either:
        response = self.job.get_job(scheduler_job_id)

        try:
            pause_status = response['settings']['schedule']['pause_status']
        except KeyError as e:
            print(f"No pause_status found in job {scheduler_job_id} settings")
            return Either(left=e)

        if pause_status != 'Paused':
            updated_settings = dict(response)
            updated_settings['settings']['schedule']['pause_status'] = 'Paused'
            self.job_client.update_job(scheduler_job_id, new_settings=updated_settings)
            print(f"Job {scheduler_job_id} has been paused.")
            return Either(right=True)
        else:
            print(f"Job {scheduler_job_id} is already paused.")
            return Either(right=False)

    @retry(retry=retry_if_exception_type(HTTPError), stop=stop_after_attempt(5),
           wait=wait_exponential(multiplier=2, max=30), reraise=True)
    def reset_job(self, scheduler_job_id: str, update_request: Dict[str, str]):
        new_settings = {'job_id': scheduler_job_id, 'new_settings': update_request}

        self.job.reset_job(new_settings)

    def patch_job_acl(self, scheduler_job_id: str, acl):
        print(f'patching job acl {scheduler_job_id} and acl {acl}')
        self.permission.patch_job(scheduler_job_id, acl)

    @retry(retry=retry_if_exception_type(HTTPError), stop=stop_after_attempt(5),
           wait=wait_exponential(multiplier=2, max=30), reraise=True)
    def find_job(self, job_name) -> Optional[str]:
        try:
            jobs = self.job.list_jobs(name=job_name)['jobs']
            if jobs is not None:
                if len(jobs) == 1:
                    return jobs[0]['job_id']
                else:
                    raise DuplicateJobNameException(f"Found more than one job with name {job_name}")
            return None
        except Exception:
            return None


class PermissionsApi(object):
    def __init__(self, client: ApiClient):
        self.client = client

    def patch_job(self, scheduler_job_id: str, data):
        return self.client.perform_query(
            'PATCH', f'/2.0/preview/permissions/jobs/{scheduler_job_id}', data=data)



# File defining all the constants being used.


SCALA_LANGUAGE = "scala"
PYTHON_LANGUAGE = "python"

DATABRICKS_HOST = "DATABRICKS_HOST"
DATABRICKS_TOKEN = "DATABRICKS_TOKEN"

PROJECT_ID_PLACEHOLDER_REGEX = "__PROJECT_ID_PLACEHOLDER__"
PROPHECY_URL_PLACEHOLDER_REGEX = "__PROPHECY_URL_PLACEHOLDER__"
PROJECT_RELEASE_VERSION_PLACEHOLDER_REGEX = "__PROJECT_RELEASE_VERSION_PLACEHOLDER__"
PROJECT_RELEASE_TAG_PLACEHOLDER_REGEX = "__PROJECT_FULL_RELEASE_TAG_PLACEHOLDER__"
PROJECT_URL_PLACEHOLDER_REGEX = "__PROPHECY_URL_PLACEHOLDER__"

PIPELINE_ID = "PipelineId"

PBT_FILE_NAME = "pbt_project.yml"
LANGUAGE = "language"
JOBS = "jobs"
PIPELINES = "pipelines"
GEMS = "gems"
PIPELINE_CONFIGURATIONS = "pipelineConfigurations"
BASE_PIPELINE = "basePipeline"
CONFIGURATIONS = "configurations"

JSON_EXTENSION = ".json"

DBT_COMPONENT = "DBTComponent"
SCRIPT_COMPONENT = "ScriptComponent"
PIPELINE_COMPONENT = "PipelineComponent"
COMPONENTS_LITERAL = "components"

FABRIC_ID = "fabric_id"
FABRIC_UID = "fabricUID"
SECRET_SCOPE = "secret_scope"

DBFS_FILE_STORE = "dbfs:/FileStore"
S3_FILE_STORE = "s3://"
PROPHECY_ARTIFACTS = "prophecy/artifacts"

NEW_JOB_STATE_FILE = "new_state.yml"



from click.testing import CliRunner
from src.pbt import build
import os

PROJECT_PATH = str(os.getcwd()) + "/test/resources/HelloWorld"


def test_build_path_default():
    runner = CliRunner()
    result = runner.invoke(build, ["--path", PROJECT_PATH])
    assert result.exit_code == 0
    assert "Found 4 pipelines" in result.output
    assert "Building 4 pipelines" in result.output
    assert "Building pipeline pipelines/customers_orders" in result.output
    assert "Building pipeline pipelines/join_agg_sort" in result.output
    assert "Building pipeline pipelines/report_top_customers" in result.output
    assert "Building pipeline pipelines/farmers-markets-irs" in result.output


def test_build_path_pipeline_filter():
    runner = CliRunner()
    result = runner.invoke(build, ["--path", PROJECT_PATH, "--pipelines", "customers_orders,join_agg_sort"])
    assert result.exit_code == 0
    assert "Found 4 pipelines" in result.output
    assert "Building 2 pipelines" in result.output
    assert "Filtering pipelines: ['customers_orders', 'join_agg_sort']" in result.output
    assert "Building pipeline pipelines/customers_orders" in result.output
    assert "Building pipeline pipelines/join_agg_sort" in result.output


def test_build_path_pipeline_with_invalid_filter():
    runner = CliRunner()
    result = runner.invoke(
        build,
        [
            "--path",
            PROJECT_PATH,
            "--pipelines",
            "customers_orders,INVALID_PIPELINE_NAME",
        ],
    )
    assert result.exit_code == 0
    assert "Found 4 pipelines" in result.output
    assert "Building 1 pipelines" in result.output
    assert "Filtering pipelines: ['customers_orders', 'INVALID_PIPELINE_NAME']" in result.output
    assert "Building pipeline pipelines/customers_orders" in result.output


def test_build_path_pipeline_invalid_filter_only():
    runner = CliRunner()
    result = runner.invoke(build, ["--path", PROJECT_PATH, "--pipelines", "INVALID_PIPELINE_NAME"])
    assert result.exit_code == 1
    assert "Found 4 pipelines" in result.output
    assert "No matching pipelines found for given pipelines names: ['INVALID_PIPELINE_NAME']" in result.output



import base64
import datetime
import json
import re
from abc import ABC

import boto3
import requests
from tenacity import retry_if_exception_type, stop_after_attempt, wait_fixed, retry

from . import AirflowRestClient
from ...utils.exceptions import DagNotAvailableException, DagFileDeletionFailedException, DagUploadFailedException, DagListParsingFailedException
from ...utils.project_models import DAG


class MWAARestClient(AirflowRestClient, ABC):

    def __init__(self, environment_name: str, region: str, access_key: str, secret_key: str):
        self.environment_name = environment_name
        self.aws_s3_client = boto3.client(
            "s3",
            region_name=region,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key
        )

        self.mwaa_client = boto3.client("mwaa", region_name=region, aws_access_key_id=access_key,
                                        aws_secret_access_key=secret_key)

        self.environment = self.mwaa_client.get_environment(Name=environment_name)
        self.dag_s3_path = self.environment['Environment']['DagS3Path']
        self._source_bucket = self._extract_bucket_name_from_arn()

    def create_secret(self, key: str, value: str) -> bool:
        return True

    def delete_dag_file(self, dag_id: str) -> bool:
        relative_path = f"{self.dag_s3_path}/{dag_id}.zip"
        try:
            self.aws_s3_client.delete_object(Bucket=self._source_bucket, Key=relative_path)
            try:
                self._get_response("dags delete {dag_id} --yes")
                return True
            except Exception:
                return False
        except Exception as e:
            raise DagFileDeletionFailedException(
                f"Error deleting file {relative_path} from bucket {self._source_bucket}", e)

    # todo improve both pause and unpause.
    def pause_dag(self, dag_id: str) -> DAG:
        response_as_text = self._get_response(f"dags pause {dag_id}")
        response_as_json = self._clean_response(response_as_text)
        return DAG(**response_as_json)

    def unpause_dag(self, dag_id: str) -> DAG:
        self.get_dag(dag_id)
        response_as_text = self._get_response(f"dags unpause {dag_id}")
        response_as_json = self._clean_response(response_as_text)
        return DAG.create_from_mwaa(response_as_json)

    def upload_dag(self, dag_id: str, file_path: str):
        relative_path = f"{self.dag_s3_path}/{dag_id}.zip"
        try:
            self.aws_s3_client.upload_file(file_path, self._source_bucket, relative_path)
        except Exception as e:
            print(f"Error uploading file {file_path} to bucket {self._source_bucket}", e)
            raise DagUploadFailedException(
                f"Error uploading file {file_path} to bucket {self._source_bucket}", e)

    @retry(retry=retry_if_exception_type(DagNotAvailableException), stop=stop_after_attempt(20),
           wait=wait_fixed(15),
           reraise=True)
    def get_dag(self, dag_id: str) -> DAG:
        response = self._get_response("dags list -o json")
        dag_list = self._extract_and_load_list_json(response)
        dag = next((dag for dag in dag_list if dag['dag_id'] == dag_id and dag['paused'] is not None), None)

        if dag is not None:
            return DAG.create_from_mwaa(dag)
        else:
            raise DagNotAvailableException(f"Dag {dag_id} not found")

    def delete_dag(self, dag_id: str) -> str:
        response = self._get_response(f"dags delete {dag_id} --yes")
        return response

    def _create_cli_token(self):
        return self.mwaa_client.create_cli_token(Name=self.environment_name)

    def _expiry_cli_token(self):
        return ExpiringValue(lambda: self._create_cli_token(), 60 * 60 * 24)

    def _headers(self):
        return {
            "Authorization": f"Bearer {self._expiry_cli_token().get_value()['CliToken']}",
            "User-Agent": "Prophecy"
        }

    def _clean_response(self, text):
        try:
            return json.loads(text)  # Try to parse the text as JSON
        except json.JSONDecodeError:
            # If unsuccessful, perform string transformations and return the result
            cleaned_text = "{\"" + text.strip() \
                .replace(", ", "\", \"") \
                .replace(": ", "\":\"") \
                .replace(",\"", ", \"") + "\"}"
            return json.loads(cleaned_text)

    def _execute_airflow_command(self, command: str):
        url = f"https://{self._expiry_cli_token().get_value()['WebServerHostname']}/aws_mwaa/cli"
        response = requests.post(url, headers=self._headers(), data=command)
        response.raise_for_status()  # Raise an HTTPError if the status is 4xx, 5xx
        return response.text  # Get the response body as text

    def _get_response(self, command: str):
        response_body = self._execute_airflow_command(command)

        decoded_response = json.loads(response_body)

        if decoded_response.get('stdout', None) is not None:
            return base64.b64decode(decoded_response['stdout']).decode('utf-8')
        else:
            raise Exception(base64.b64decode(decoded_response['stderr']).decode('utf-8'))

    def _extract_bucket_name_from_arn(self) -> str:
        s3_arn_regex = r"arn:aws:s3:::(.+)"

        match = re.match(s3_arn_regex, self.environment['Environment']['SourceBucketArn'])
        if match:
            return match.group(1)  # Return the captured group
        else:
            raise ValueError(f"No match found for {self.dag_s3_path}")

    def _extract_and_load_list_json(self, data):
        # Using regular expression to find the JSON part
        match = re.search(r'\[.*\]', data)
        if match:
            json_part = match.group()
            try:
                # Parsing the JSON part
                return json.loads(json_part)
            except json.JSONDecodeError as e:
                print(f"Error decoding JSON: {e}")
                raise DagListParsingFailedException(f"Error decoding JSON: {e}", e)
        else:
            print("No JSON found in the data")
            raise DagListParsingFailedException("No JSON data found in dag listing")


class ExpiringValue:
    def __init__(self, generate_value, duration):
        self.generate_value = generate_value
        self.duration = duration
        self.value = None
        self.expiration_time = datetime.datetime.now() + datetime.timedelta(seconds=duration)

    def get_value(self):
        if self.value is None or datetime.datetime.now() > self.expiration_time:
            self.value = self.generate_value()
            self.expiration_time = datetime.datetime.now() + datetime.timedelta(seconds=self.duration)
        if self.value is None:
            raise Exception("Failed to generate value")
        return self.value

    def is_valid(self):
        return datetime.datetime.now() < self.expiration_time



from abc import ABC, abstractmethod

from ...utils.project_models import DAG


class AirflowRestClient(ABC):

    @abstractmethod
    def delete_dag_file(self, dag_id: str):
        pass

    @abstractmethod
    def pause_dag(self, dag_id: str) -> DAG:
        pass

    @abstractmethod
    def unpause_dag(self, dag_id: str) -> DAG:
        pass

    @abstractmethod
    def create_secret(self, key: str, value: str) -> bool:
        pass

    @abstractmethod
    def upload_dag(self, dag_id: str, file_path: str):
        pass

    @abstractmethod
    def get_dag(self, dag_id: str) -> DAG:
        pass

    @abstractmethod
    def delete_dag(self, dag_id: str) -> str:
        pass



import copy
import os
from typing import List

import yaml

from .gems import GemsDeployment
from ..deployment.jobs.airflow import AirflowJobDeployment, AirflowGitSecrets, EMRPipelineConfigurations, \
    DataprocPipelineConfigurations
from ..deployment.jobs.databricks import DatabricksJobsDeployment, ScriptComponents, PipelineConfigurations, \
    DBTComponents
from ..deployment.pipeline import PipelineDeployment
from ..entities.project import Project
from ..utility import custom_print as log, Either, is_online_mode
from ..utility import remove_null_items_recursively
from ..utils.constants import NEW_JOB_STATE_FILE
from ..utils.project_config import ProjectConfig
from ..utils.project_models import StepMetadata, Operation, StepType, Status, Colors


class ProjectDeployment:
    def __init__(self, project: Project, project_config: ProjectConfig):
        self.project = project
        self.project_config = project_config

        self._databricks_jobs = DatabricksJobsDeployment(project, project_config)
        self._airflow_jobs = AirflowJobDeployment(project, project_config)

        self._script_component = ScriptComponents(project, self._databricks_jobs, project_config)
        self._pipeline_configurations = PipelineConfigurations(project, self._databricks_jobs,
                                                               project_config)
        self._emr_pipeline_configurations = EMRPipelineConfigurations(project, self._airflow_jobs, project_config)

        self._dataproc_pipeline_configurations = DataprocPipelineConfigurations(project, self._airflow_jobs,
                                                                                project_config)

        self._dbt_component = DBTComponents(project, self._databricks_jobs, project_config)
        self._airflow_git_secrets = AirflowGitSecrets(project, self._airflow_jobs, project_config)

        self._pipelines = PipelineDeployment(project, self._databricks_jobs, self._airflow_jobs,
                                             project_config)

        # add gems Deployment.
        self._gems = GemsDeployment(project, project_config)

    def headers(self):
        summary = self._gems.summary() + self._script_component.summary() + \
                  self._dbt_component.summary() + self._airflow_git_secrets.summary() + \
                  self._pipeline_configurations.summary() + self._emr_pipeline_configurations.summary() + \
                  self._dataproc_pipeline_configurations.summary() + self._pipelines.summary() + \
                  self._databricks_jobs.summary() + self._airflow_jobs.summary()

        if len(summary) == 0:
            summary = ["No Job and pipelines to build"]

        summary_header = [StepMetadata("Summary", "Summary", Operation.Build, StepType.Summary)]

        header_components = (
            summary_header,
            self._gems.headers(),
            self._script_component.headers(),
            self._dbt_component.headers(),
            self._airflow_git_secrets.headers(),
            self._pipeline_configurations.headers(),
            self._emr_pipeline_configurations.headers(),
            self._dataproc_pipeline_configurations.headers(),
            self._pipelines.headers(),
            self._databricks_jobs.headers(),
            self._airflow_jobs.headers()
        )

        headers = sum(header_components, [])

        # 1st steps have to be summary
        for header in headers:
            log(step_id=header.id, step_metadata=header)

        for step in summary:
            log(message=step, step_id="Summary")

        log(step_status=Status.SUCCEEDED, step_id="Summary")

        return headers

    def build(self, pipelines, ignore_build_errors, ignore_parse_errors):
        self._pipelines.build(pipelines, ignore_build_errors, ignore_parse_errors)

    def validate(self, treat_warning_as_errors):
        self._pipelines.validate(treat_warning_as_errors)

    def test(self, ):
        self._pipelines.test()

    def _deploy_gems(self):
        gems_responses = self._gems.deploy()

        if gems_responses is not None and any(response.is_left for response in gems_responses):
            raise Exception("Gems deployment failed.")

    def _deploy_scripts(self):
        script_responses = self._script_component.deploy()

        if script_responses is not None and any(response.is_left for response in script_responses):
            raise Exception("Script deployment failed.")

    def _deploy_dbt_components(self):
        dbt_responses = self._dbt_component.deploy()

        if dbt_responses is not None and any(response.is_left for response in dbt_responses):
            raise Exception("DBT deployment failed.")

    def _deploy_airflow_git_secrets(self):
        airflow_git_secrets_responses = self._airflow_git_secrets.deploy()

        if airflow_git_secrets_responses is not None and any(
                response.is_left for response in airflow_git_secrets_responses):
            raise Exception("Airflow git secrets deployment failed.")

    def _deploy_pipeline_configs(self):
        pipeline_config_responses = self._pipeline_configurations.deploy()

        if pipeline_config_responses is not None and any(response.is_left for response in pipeline_config_responses):
            raise Exception("Pipeline config deployment failed.")

    def _deploy_emr_pipeline_config(self):
        emr_pipeline_config_responses = self._emr_pipeline_configurations.deploy()

        if emr_pipeline_config_responses is not None and any(
                response.is_left for response in emr_pipeline_config_responses):
            raise Exception("EMR pipeline config deployment failed.")

    def _deploy_dataproc_pipeline_config(self):
        dataproc_pipeline_config_responses = self._dataproc_pipeline_configurations.deploy()

        if dataproc_pipeline_config_responses is not None and any(
                response.is_left for response in dataproc_pipeline_config_responses):
            raise Exception("Dataproc pipeline config deployment failed.")

    def _deploy_pipelines(self):
        pipeline_responses = self._pipelines.deploy()

        if pipeline_responses is not None and any(response.is_left for response in pipeline_responses):
            raise Exception("Pipeline deployment failed.")

    def _deploy_databricks_jobs(self) -> List[Either]:
        databricks_jobs_responses = self._databricks_jobs.deploy()

        return databricks_jobs_responses

    def _deploy_airflow_jobs(self) -> List[Either]:
        airflow_jobs_responses = self._airflow_jobs.deploy()

        return airflow_jobs_responses

    def deploy(self, job_ids):
        if is_online_mode():
            self._deploy_gems()

        self._deploy_scripts()
        self._deploy_dbt_components()
        self._deploy_airflow_git_secrets()
        self._deploy_pipeline_configs()
        self._deploy_emr_pipeline_config()
        self._deploy_dataproc_pipeline_config()

        if not self.project_config.skip_builds:
            self._deploy_pipelines()
        else:
            log("\nSkipping pipeline deployment as skip_builds is set to true.\n")

        databricks_responses = self._deploy_databricks_jobs()
        airflow_responses = self._deploy_airflow_jobs()

        new_state_config = copy.deepcopy(self.project_config.jobs_state)

        # only jobs changes state_config.

        new_state_config.update_state(databricks_responses + airflow_responses)

        base_path = os.path.join(os.getcwd(), ".pbt")

        if not os.path.exists(base_path):
            os.makedirs(base_path)

        if self.project_config.based_on_file or self.project_config.migrate:
            path = os.path.join(base_path, NEW_JOB_STATE_FILE)
            yaml_str = yaml.dump(data=remove_null_items_recursively(new_state_config.dict()))

            with open(path, 'w') as file:
                file.write(yaml_str)

        # Only fail when there is a failure in jobs deployment.
        if databricks_responses is not None and any(response.is_left for response in databricks_responses):
            for response in databricks_responses:
                if response.is_left:
                    print(response.left)
            raise Exception("Databricks jobs deployment failed.")

        if airflow_responses is not None and any(response.is_left for response in airflow_responses):
            raise Exception("Airflow jobs deployment failed.")

        log(f"\n\n{Colors.OKCYAN}Deployment completed successfully.\n{Colors.ENDC}")

        if self.project_config.migrate:
            log(f"\n\n{Colors.OKCYAN}Migrating project to new version. \n{Colors.ENDC}")
            try:
                fabric_config_str = yaml.dump(
                    data=remove_null_items_recursively(self.project_config.fabric_config.dict()))
                system_config_str = yaml.dump(
                    data=remove_null_items_recursively(self.project_config.system_config.dict()))
                config_override_str = yaml.dump(
                    data=remove_null_items_recursively(self.project_config.configs_override.dict()))

                with open(os.path.join(base_path, "fabrics.yml"), 'w') as file:
                    file.write(fabric_config_str)

                with open(os.path.join(base_path, "system.yml"), 'w') as file:
                    file.write(system_config_str)

                with open(os.path.join(base_path, "override.yml"), 'w') as file:
                    file.write(config_override_str)

                log(f"\n\n{Colors.OKCYAN}Successfully migrated project to new version.\n{Colors.ENDC}")
            except Exception as e:
                log(f"\n\n{Colors.FAIL}Failed to migrate project to new version.\n{Colors.ENDC}", e)
                if os.path.isdir(base_path):
                    os.rmdir(base_path)

        return databricks_responses + airflow_responses



import os
from typing import Optional

from .deployment.project import ProjectDeployment
from .entities.project import Project
from .utils.project_config import ProjectConfig


class PBTCli(object):
    """Command line interface for PBT."""

    def __init__(self, project: Project, project_config: ProjectConfig):
        self.project = ProjectDeployment(project, project_config)

    def headers(self):
        """Print headers."""
        self.project.headers()

    def deploy(self, job_ids):
        """Deploy pipelines."""
        self.project.deploy(job_ids)

    @classmethod
    def from_conf_folder(cls, project_path: str, project_id: str = "", conf_folder: str = "",
                         release_tag: Optional[str] = "",
                         release_version: str = "", fabric_ids: str = "", job_ids: str = "", skip_builds: bool = False,
                         dependant_project_paths: str = "", migrate: bool = False):
        """Create PBTCli from conf folder."""
        project = Project(project_path, project_id, release_tag, release_version, dependant_project_paths)
        project_config = ProjectConfig.from_conf_folder(project, conf_folder, fabric_ids, job_ids, skip_builds, migrate)
        return cls(project, project_config)

    def build(self, pipelines, ignore_build_errors, ignore_parse_errors):
        self.project.build(pipelines, ignore_build_errors, ignore_parse_errors)

    def test(self, driver_library_path: str):
        if self.project.project.project_language == "PYTHON":
            os.environ["SPARK_JARS_CONFIG"] = driver_library_path if driver_library_path else ""
        self.project.test()

    def validate(self, treat_warnings_as_errors: bool):
        self.project.validate(treat_warnings_as_errors)



import json
import os
import re
import tempfile
from abc import ABC
from pathlib import Path
from typing import Optional

import requests
from google.auth.transport.requests import AuthorizedSession
from google.cloud import secretmanager, storage
from google.cloud.storage import Blob
from google.oauth2 import service_account
from google.oauth2.gdch_credentials import ServiceAccountCredentials
from requests import HTTPError
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed

from . import AirflowRestClient
from ...utils.exceptions import DagUploadFailedException, DagFileDeletionFailedException
from ...utils.project_models import DAG

pattern = re.compile(r"^g[c]?s://([a-z0-9][-a-z0-9.]*[a-z0-9]):?/(.*)?$")


class GCSPathInfo:

    def __init__(self, bucket: str, path: Optional[str]):
        self.bucket = bucket
        self.path = path

    @staticmethod
    def get_gcs_path_info(dag_location: str):
        match = pattern.match(dag_location)

        if match:
            bucket, path = match.groups()
            gcs_path_info = GCSPathInfo(bucket.strip("/"), "/".join(path.split("/")).rstrip("/"))
        else:
            gcs_path_info = None

        return gcs_path_info


class ComposerRestClient(AirflowRestClient, ABC):
    _SCOPES = ["https://www.googleapis.com/auth/cloud-platform"]
    _CONNECTIONS_SECRET_PREFIX = "airflow-connections-"
    _pattern = re.compile(pattern)
    _headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}

    def __init__(self, airflow_url: str, project_id: str, client_id: Optional[str], key_json: str, dag_location: str,
                 location: Optional[str] = None):
        self.airflow_url = airflow_url
        self.project_id = project_id
        self.client_id = client_id
        self.key_json_file = key_json
        self.dag_location = dag_location
        self.location = location

        self.file_path = self._create_key_json_file(key_json)

        self.storage_handler = storage.Client.from_service_account_info(json.loads(key_json), project=project_id)

    def delete_dag_file(self, dag_id: str):
        gcs_path_info = GCSPathInfo.get_gcs_path_info(self.dag_location)
        blob: Blob = Blob(self._get_dag_location(gcs_path_info, dag_id), gcs_path_info.bucket)
        try:
            self.storage_handler.bucket(gcs_path_info.bucket).delete_blob(blob.name)
        except Exception as e:
            raise DagFileDeletionFailedException(f"Failed to delete DAG file: {blob.name} ", e)

    def pause_dag(self, dag_id: str) -> DAG:
        return self._set_pause_state(dag_id, True)

    def upload_dag(self, dag_id: str, file_path: str):
        gcs_path_info = GCSPathInfo.get_gcs_path_info(self.dag_location)
        bucket = self.storage_handler.get_bucket(gcs_path_info.bucket)
        dag_location = self._get_dag_location(gcs_path_info, dag_id)

        blob: Blob = Blob(dag_location, gcs_path_info.bucket)

        try:
            bucket.blob(blob_name=blob.name).upload_from_filename(file_path)
        except Exception as e:
            raise DagUploadFailedException(f"Failed to upload DAG file: {dag_location} ", e)
        finally:
            if Path(file_path).exists():
                os.remove(file_path)

    def unpause_dag(self, dag_id: str) -> DAG:
        return self._set_pause_state(dag_id, False)

    # todo maybe we can add labmda function to print in right format
    def create_secret(self, key: str, value: str) -> bool:
        client = None
        try:
            client = self._get_secrets_manager()
            secret_name = f"projects/{self.project_id}/secrets/{key}"

            try:
                client.get_secret(name=secret_name)
                client.add_secret_version(
                    request={"parent": secret_name, "payload": {"data": value.encode('UTF-8')}}
                )
            except Exception as e:
                print(f"Failed to get secret: {secret_name}, trying to create default one", e)

                parent = f"projects/{self.project_id}"
                secret = client.create_secret(
                    request={
                        "parent": parent,
                        "secret_id": key,
                        "secret": {
                            "replication": {"automatic": {}}
                        },
                    }
                )
                client.add_secret_version(request={
                    "parent": secret.name,
                    "payload": value.encode('UTF-8'),
                })

        except Exception as e:
            print("Failed to create Secret manager client", e)
            return False
        finally:
            if client is not None:
                client.__exit__()

        return True

    def get_dag(self, dag_id: str) -> DAG:
        response = requests.request('GET', f"{self.airflow_url}/api/v1/dags/{dag_id}",
                                    headers=self._headers)
        response.raise_for_status()
        response_data = response.json()

        # Parse JSON into Dag class
        dag = DAG(**response_data)
        return dag

    def delete_dag(self, dag_id: str) -> str:
        response = requests.request('DELETE', f"{self.airflow_url}/api/v1/dags/{dag_id}")
        response.raise_for_status()
        return response.text

    def _create_key_json_file(self, key_json: str) -> str:
        temp_file = tempfile.NamedTemporaryFile(prefix="composer", suffix=".json", delete=False)

        # Write content to the temporary file
        Path(temp_file.name).write_text(key_json)

        return temp_file.name

    def _get_dag_location(self, gcp_path_info: GCSPathInfo, dag_id: str) -> str:
        if gcp_path_info is not None and gcp_path_info.path is not None:
            return f"{gcp_path_info.path}/{dag_id}.zip"
        else:
            print(f"Shouldn't be here: {gcp_path_info}")
            return f"{dag_id}.zip"

    def _get_secrets_manager(self):
        # fill up how to get the credentials
        credentials = ServiceAccountCredentials.from_service_account_file(self.key_json_file)
        client = secretmanager.SecretManagerServiceClient(credentials=credentials)
        return client

    def _get_authenticated_session(self):

        if self.client_id is not None and len(self.client_id) > 0:
            print(f'Client Id {self.client_id}')
            credentials = service_account.IDTokenCredentials \
                .from_service_account_info(json.loads(self.key_json_file)) \
                .with_target_audience(self.client_id) \
                .with_scopes(self._SCOPES)

        else:
            credentials = service_account.Credentials.from_service_account_info(json.loads(self.key_json_file)) \
                .with_scopes(self._SCOPES)
        return AuthorizedSession(credentials)

    @retry(retry=retry_if_exception_type(HTTPError), stop=stop_after_attempt(20), wait=wait_fixed(15), reraise=True)
    def _set_pause_state(self, dag_id: str, is_paused: bool) -> DAG:
        session = self._get_authenticated_session()
        response = session.request(method='PATCH', url=f"{self.airflow_url}/api/v1/dags/{dag_id}",
                                   data=json.dumps({"is_paused": is_paused}),
                                   headers=self._headers)
        response.raise_for_status()
        response_data = response.json()
        return DAG.create(response_data)

    def put_object_from_file(self, bucket: str, key: str, file_path: str):
        try:
            self.storage_handler.bucket(bucket).blob(key).upload_from_filename(file_path)
        except Exception as e:
            raise DagUploadFailedException(f"Failed to upload file: {file_path} ", e)

    def put_object(self, bucket: str, key: str, content: str):
        try:
            self.storage_handler.bucket(bucket).blob(key).upload_from_string(content)
        except Exception as e:
            raise DagUploadFailedException(f"Failed to upload content to bucket: {bucket} ", e)



import base64
import binascii
import hashlib
import os
import zipfile
from abc import ABC
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, Optional, List

import yaml

from .databricks import get_fabric_label
from ...client.airflow.airflow_utility import create_airflow_client, get_fabric_provider_type
from ...client.rest_client_factory import RestClientFactory
from ...deployment import JobInfoAndOperation, OperationType, JobData, EntityIdToFabricId
from ...entities.project import Project
from ...utility import custom_print as log, Either
from ...utils.constants import FABRIC_UID
from ...utils.project_config import JobInfo, ProjectConfig, FabricInfo, update_state, await_futures_and_update_states
from ...utils.project_models import StepMetadata, Operation, StepType, Colors


def generate_secure_content(content: str, salt: str) -> str:
    iterations = 10000
    key_length = 128

    password = content.encode('utf-8')
    salt_bytes = salt.encode('utf-8')

    derived_key = hashlib.pbkdf2_hmac('sha256', password, salt_bytes, iterations, dklen=key_length // 8)
    hash_bytes = binascii.hexlify(derived_key)

    return base64.b64encode(hash_bytes).decode('utf-8').replace('\\W+', '_')


def calculate_checksum(input_str, salt=None):
    salt = salt or "better_salt_than_never"

    md = hashlib.sha256()
    input_bytes = input_str.encode('utf-8')
    salt_bytes = salt.encode('utf-8')

    md.update(input_bytes)
    md.update(salt_bytes)

    digest_bytes = md.digest()

    return ''.join(f'{byte:02x}' for byte in digest_bytes)


def does_dag_file_exist(rdc: Dict[str, str]) -> bool:
    return rdc is not None and 'dag.py' in rdc


def filter_job_files(rdc: Dict[str, str]):
    filtered_files = {file_name: file_content for file_name, file_content in rdc.items()
                      if
                      file_name == 'dag.py' or '__init__.py' or file_name == 'prophecy-job.json' or
                      'tasks/' in file_name}
    k = sorted(filtered_files.items())
    return k


def sanitize_job(job_id) -> str:
    return ('_' if job_id is None else job_id).split('/')[-1].replace('\\W', '_')


def get_zipped_dag_name(dag_name: str):
    return f"/tmp/{dag_name}.zip"


def zip_folder(rdc: dict, output_path):
    with zipfile.ZipFile(output_path, 'w', zipfile.ZIP_STORED) as zipf:
        for path, content in rdc.items():
            zipf.writestr(path, content)


class AirflowJob(JobData, ABC):

    def __init__(self, job_pbt: dict, prophecy_job_yaml: str, rdc: Dict[str, str], rdc_with_placeholder: Dict[str, str],
                 sha: Optional[str], fabric_override: Optional[str] = None):
        self.job_pbt = job_pbt
        self.prophecy_job_yaml = prophecy_job_yaml
        self.rdc = rdc
        self.rdc_with_placeholder = rdc_with_placeholder
        self.sha = sha
        self.fabric_override = fabric_override

        self.prophecy_job_json_dict = self._initialize_prophecy_job_json()

    @property
    def name(self):
        return self.job_pbt.get('name', None)

    @property
    def is_valid_job(self):
        prophecy_job_yaml_dict = self.prophecy_job_json_dict

        return self.job_pbt is not None and self.prophecy_job_yaml is not None and self.rdc is not None and \
            prophecy_job_yaml_dict.get('metainfo', {}).get('fabricId', None) is not None

    # we can't use pbt file because it doesn't have fabric per pipeline which airflow jobs supports
    @property
    def pipeline_and_fabric_ids(self) -> List[EntityIdToFabricId]:
        pipeline_and_fabric_ids = []
        prophecy_job_yaml_dict = self.prophecy_job_json_dict

        if self.job_pbt is not None and self.prophecy_job_yaml is not None and self.rdc is not None:
            for process_id, process in prophecy_job_yaml_dict.get('processes', {}).items():
                properties = process.get('properties', {})

                cluster_size = properties.get('clusterSize', None)
                pipeline_id = properties.get('pipelineId', {})

                if cluster_size is not None:
                    # shady stuff but can't help
                    pipeline_and_fabric_ids.append(EntityIdToFabricId(pipeline_id, cluster_size.split('/')[0]))

        return pipeline_and_fabric_ids

    def validate_prophecy_managed_checksum(self, salt: str):
        file_joiner: str = "$$$"
        # always use rdc_with_placeholder to calculate checksum
        content = file_joiner.join(content for (name, content) in filter_job_files(self.rdc_with_placeholder))
        checksum = calculate_checksum(content, salt)
        return self.sha == checksum

    @property
    def job_files(self):
        return filter_job_files(self.rdc)

    @property
    def dag_name(self):
        return self.prophecy_job_json_dict.get('metainfo', {}).get('dagName', None)

    @property
    def is_disabled(self):
        return self.prophecy_job_json_dict.get("metainfo", {}).get('enabled', False) is False

    @property
    def is_enabled(self):
        return self.prophecy_job_json_dict.get("metainfo", {}).get('enabled', False) is True

    @property
    def fabric_id(self):
        if self.fabric_override is not None and len(self.fabric_override) > 0:
            return self.fabric_override
        else:
            fabric_id = self.job_pbt.get(FABRIC_UID)
            return str(fabric_id) if fabric_id is not None else None

    @property
    def pipelines(self):
        return self.job_pbt.get('pipelines', [])

    @property
    def has_dbt_component(self):
        return any(value.get('component', None) == 'dbt' for value in self.prophecy_job_json_dict['processes'].values())

    def _initialize_prophecy_job_json(self) -> dict:
        try:
            return yaml.unsafe_load(self.prophecy_job_yaml)
        except Exception as e:
            log("Error while loading prophecy job yaml", e)
            return {}


class AirflowJobDeployment:
    _REMOVE_JOBS_STEP_ID = "remove-airflow-jobs"
    _ADD_JOBS_STEP_ID = "add-airflow-jobs"
    _PAUSE_JOBS_STEP_ID = "pause-airflow-jobs"
    _RENAME_JOBS_STEP_ID = "rename-airflow-jobs"
    _SKIP_JOBS_STEP_ID = "skip-airflow-jobs"

    def __init__(self, project: Project, project_config: ProjectConfig):
        self._project = project
        self._project_config = project_config
        self._jobs_state = project_config.jobs_state
        self._fabrics_config = project_config.fabric_config

        self._airflow_clients = {}
        self.deployment_run_override_config = project_config.configs_override
        self._rest_client_factory = RestClientFactory.get_instance(RestClientFactory, self._fabrics_config)
        self._airflow_jobs: Dict[str, AirflowJob] = self._initialize_airflow_jobs()

        (self.valid_airflow_jobs, self._invalid_airflow_jobs,
         self._airflow_jobs_without_code) = self._initialize_valid_airflow_jobs()

        self.prophecy_managed_dbt_jobs: Dict[str, AirflowJob] = self._initialize_prophecy_managed_dbt_jobs()

    def get_airflow_client(self, fabric_id):
        return self._rest_client_factory.airflow_client(fabric_id)

    def summary(self):
        summary = []

        summary.extend([f"Airflow job to remove {job.id}" for job in self._list_remove_jobs()])
        summary.extend([f"Airflow job to add {job}" for job in self._add_jobs().keys()])
        summary.extend([f"Airflow job to pause {job}" for job in self._pause_jobs().keys()])
        summary.extend([f"Airflow job to rename {job.id}" for job in self._rename_jobs()])
        summary.extend([f"Airflow job to skip {job}" for job in self._skip_jobs().keys()])

        return summary

    @property
    def _operation_to_step_id(self):
        return {
            Operation.Remove: self._REMOVE_JOBS_STEP_ID,
            Operation.Add: self._ADD_JOBS_STEP_ID,
            Operation.Pause: self._PAUSE_JOBS_STEP_ID,
            Operation.Rename: self._RENAME_JOBS_STEP_ID,
            Operation.Skipped: self._SKIP_JOBS_STEP_ID
        }

    def headers(self):
        job_types = {
            Operation.Remove: self._list_remove_jobs(),
            Operation.Add: self._add_jobs(),
            Operation.Pause: self._pause_jobs(),
            Operation.Rename: self._rename_jobs(),
            Operation.Skipped: self._skip_jobs()
        }

        all_headers = []

        for job_action, jobs in job_types.items():
            if jobs:
                len_jobs = len(jobs)
                job_header_suffix = f'{len_jobs} Airflow job' if len_jobs == 1 else f'{len_jobs} Airflow jobs'
                step_id = self._operation_to_step_id[job_action]
                all_headers.append(
                    StepMetadata(step_id, f"{job_action.value} {job_header_suffix}",
                                 job_action, StepType.Job)
                )

        return all_headers

    def deploy(self):
        if len(self.headers()) > 0:
            log(f"{Colors.OKBLUE}\n\nDeploying airflow jobs{Colors.ENDC}\n")

        responses = self._deploy_remove_jobs() + self._deploy_pause_jobs() + \
            self._deploy_add_jobs() + self._deploy_rename_jobs()

        self._deploy_skipped_jobs()

        return responses

    def _initialize_airflow_jobs(self):
        jobs = {}

        for job_id, parsed_job in self._project.jobs.items():

            fabric_override = self.deployment_run_override_config.find_fabric_override_for_job(job_id)
            fabric_override = str(fabric_override) if fabric_override is not None else None

            job_fabric = parsed_job.get('fabricUID', None)
            job_fabric = str(job_fabric) if job_fabric is not None else None

            does_fabric_exist = self._fabrics_config.get_fabric(
                job_fabric) is not None or self._fabrics_config.get_fabric(fabric_override) is not None

            if 'Databricks' not in parsed_job.get('scheduler', None) and \
                    self.deployment_run_override_config.is_job_to_run(job_id) and does_fabric_exist:

                rdc_with_placeholders = self._project.load_airflow_folder_with_placeholder(job_id)
                rdc = self._project.load_airflow_folder(job_id)

                aspects = self._project.load_airflow_aspect(job_id)

                prophecy_job_json = None
                sha = None

                if rdc is not None and 'prophecy-job.json' in rdc:
                    prophecy_job_json = rdc.get('prophecy-job.json', None)

                if aspects is not None:
                    try:
                        sha = yaml.safe_load(aspects).get('sha', None)
                    except Exception as e:
                        log("Error while loading prophecy job yaml", e)

                jobs[job_id] = AirflowJob(parsed_job, prophecy_job_json, rdc, rdc_with_placeholders, sha,
                                          fabric_override=self.deployment_run_override_config.find_fabric_override_for_job(
                                              job_id))

        return jobs

    def _initialize_valid_airflow_jobs(self):
        valid_airflow_jobs = {}
        invalid_airflow_jobs = {}
        airflow_jobs_without_code = {}

        for job_id, job_data in self._airflow_jobs.items():
            if job_data.is_valid_job:
                response = self._validate_airflow_job(job_id, job_data)
                if response.is_right:
                    valid_airflow_jobs[job_id] = job_data

                else:
                    invalid_airflow_jobs[job_id] = job_data

            else:
                airflow_jobs_without_code[job_id] = job_data

        return valid_airflow_jobs, invalid_airflow_jobs, airflow_jobs_without_code

    def _validate_airflow_job(self, job_id: str, job_data: AirflowJob):

        is_prophecy_managed_fabric = self._fabrics_config.is_prophecy_managed(
            job_data.fabric_id)
        rdc = job_data.rdc

        if does_dag_file_exist(rdc) is False:
            return Either(
                left=Exception(f"Please open the Job `{job_id}` in editor, check diagnostic errors and release again."))

        elif is_prophecy_managed_fabric:
            if job_data.validate_prophecy_managed_checksum(self._project_config.system_config.prophecy_salt):
                return Either(right=job_data.job_files)

            else:
                return Either(left=Exception(
                    f"Job `{job_id}` has been externally edited. Please open the Job in editor and release again."))
        else:
            return Either(right=rdc)

    # we won't be able to completely validated the prophecy_job_json structure to prophecy_job.json.
    def _initialize_prophecy_managed_dbt_jobs(self) -> Dict[str, AirflowJob]:
        prophecy_managed_dbt_jobs = {}

        for job_id, job_data in self.valid_airflow_jobs.items():

            is_job_enabled = job_data.is_enabled
            is_prophecy_managed_fabric = self._fabrics_config.is_prophecy_managed(
                job_data.fabric_id)

            if is_job_enabled and is_prophecy_managed_fabric and job_data.has_dbt_component:
                prophecy_managed_dbt_jobs[job_id] = job_data

        return prophecy_managed_dbt_jobs

    def _list_remove_jobs(self) -> List[JobInfo]:
        return self._jobs_to_be_deleted() + self._jobs_with_fabric_changed()

    def _deploy_remove_jobs(self):
        futures = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            for job_info in self._list_remove_jobs():
                futures.append(executor.submit(lambda j_info=job_info: self._remove_job(j_info)))

        return await_futures_and_update_states(futures, self._operation_to_step_id[Operation.Remove])

    def _remove_job(self, job_info: JobInfo):
        client = self.get_airflow_client(job_info.fabric_id)

        try:
            client.delete_dag_file(job_info.external_job_id)
            log(f"Successfully deleted job {job_info.external_job_id} from job_id {job_info.id}",
                step_id=self._REMOVE_JOBS_STEP_ID)
            return Either(right=JobInfoAndOperation(job_info, OperationType.DELETED))

        except Exception as e:
            log(f"Error while deleting job {job_info.external_job_id} from job_id {job_info.id}, Please delete the Dag manually from Dag location if required.",
                exception=e,
                step_id=self._REMOVE_JOBS_STEP_ID)
            return Either(right=JobInfoAndOperation(job_info, OperationType.DELETED))

    '''
        Compare what exists in state config to what exists in code.
     '''

    def _jobs_to_be_deleted(self) -> List[JobInfo]:
        all_jobs = {**self.valid_airflow_jobs, **self._invalid_airflow_jobs, **self._airflow_jobs_without_code}
        return [
            airflow_job for airflow_job in self._jobs_state.airflow_jobs
            if not any(
                airflow_job.id == job_id
                for job_id in list(all_jobs.keys())  # check from available airflow jobs.
            ) and self._fabrics_config.get_fabric(airflow_job.fabric_id) is not None
        ]

    '''
        There exist a job in folder whose fabric id is different from the one in state config.
        folder can only have one job with a given job id.
    '''

    def _jobs_with_fabric_changed(self) -> List[JobInfo]:

        return [
            airflow_job for airflow_job in self._jobs_state.airflow_jobs
            if any(
                airflow_job.id == job_id and airflow_job.fabric_id != job_data.fabric_id
                for job_id, job_data in self.valid_airflow_jobs.items()  # Check only from valid airflow jobs.
            )
        ]

    def _rename_jobs(self) -> List[JobInfo]:
        return [
            airflow_job for airflow_job in self._jobs_state.airflow_jobs

            if any(
                airflow_job.id == job_id and airflow_job.fabric_id == job_data.fabric_id and
                job_data.is_enabled and airflow_job.external_job_id != job_data.dag_name
                for job_id, job_data in self.valid_airflow_jobs.items()
            )
        ]

    def _all_removed_airflow_jobs(self) -> List[JobInfo]:
        return self._rename_jobs() + self._jobs_with_fabric_changed() + self._jobs_to_be_deleted()

    # Here we are looking for the flipped state of the job.
    # jobs which are enabled in state config but disabled in code.
    # and always get jobs from deployment state.
    def _pause_jobs(self) -> Dict[str, JobInfo]:
        old_enabled_job = {job.id: job for job in self._jobs_state.airflow_jobs if job.is_paused is False}

        old_enabled_job_which_are_disabled_in_new = {
            job_id: job_info
            for job_id, job_info in old_enabled_job.items()
            if any(
                job_id == _id and job_info.fabric_id == job_data.fabric_id and job_data.is_disabled  # disabled in new
                for _id, job_data in self.valid_airflow_jobs.items()
            )
        }

        jobs_not_in_removed_jobs = {
            job_id: job_info
            for job_id, job_info in old_enabled_job_which_are_disabled_in_new.items()
            if not self._all_removed_airflow_jobs() and not any(
                airflow_job.id == job_id
                for airflow_job in self._all_removed_airflow_jobs()
            )
        }

        return jobs_not_in_removed_jobs

    # always add new jobs only if they are enabled.
    # this is equivalent to refresh because airflow merge based on it's dag_name.
    def _add_jobs(self) -> Dict[str, AirflowJob]:
        return {
            job_id: job_data for job_id, job_data in self.valid_airflow_jobs.items()
            if job_data.is_enabled is True
        }

    def _deploy_add_jobs(self):
        futures = []

        with ThreadPoolExecutor(max_workers=10) as executor:
            for job_id, job_data in self._add_jobs().items():
                futures.append(executor.submit(lambda j_id=job_id, j_data=job_data: self._add_job(j_id, j_data)))

        return await_futures_and_update_states(futures, self._operation_to_step_id[Operation.Add])

    def _add_job(self, job_id, job_data):
        dag_name = job_data.dag_name
        zipped_dag_name = get_zipped_dag_name(dag_name)
        zip_folder(self._project.load_airflow_folder(job_id), zipped_dag_name)
        client = self.get_airflow_client(fabric_id=job_data.fabric_id)

        fabric_config = self._fabrics_config.get_fabric(job_data.fabric_id)
        fabric_name = fabric_config.name if fabric_config is not None else None

        fabric_label = get_fabric_label(fabric_name, job_data.fabric_id)
        try:
            client.upload_dag(dag_name, zipped_dag_name)
            try:
                client.unpause_dag(dag_name)
                log(f"{Colors.OKGREEN}Successfully un-paused dag:{dag_name} for job:{job_id} and fabric:{fabric_label}{Colors.ENDC}",
                    step_id=self._ADD_JOBS_STEP_ID)
            except Exception as e:
                log(f"{Colors.FAIL}Failed to un-pause dag with name:{dag_name} for job:{job_id} and fabric:{fabric_label}{Colors.ENDC}",
                    exception=e, step_id=self._ADD_JOBS_STEP_ID)

                return Either(left=e)

            log(f"{Colors.OKGREEN}Successfully added job:{dag_name} for job_id:{job_id} on fabric:{fabric_label}{Colors.ENDC}",
                step_id=self._ADD_JOBS_STEP_ID)

            job_info = JobInfo.create_job(job_data.name, job_id, job_data.fabric_id, dag_name,
                                          self._project.release_tag,
                                          job_data.is_disabled,
                                          get_fabric_provider_type(job_data.fabric_id, self._project_config))

            return Either(right=JobInfoAndOperation(job_info, OperationType.CREATED))
        except Exception as e:
            log(f"{Colors.FAIL}Failed to upload_dag for job_id:{job_id} with dag_name {dag_name}{Colors.ENDC}", e,
                step_id=self._ADD_JOBS_STEP_ID)
            return Either(left=e)

    def _deploy_skipped_jobs(self):
        for job_id, message in self._skip_jobs().items():
            log(f"Skipping job_id: {job_id} encountered some error ", exception=message.left,
                step_id=self._SKIP_JOBS_STEP_ID)
        if len(self._skip_jobs()) > 0:
            update_state([Either(right=True)], self._operation_to_step_id[Operation.Skipped])

    def _skip_jobs(self):
        return {job_id: self._validate_airflow_job(job_id, job_data)
                for job_id, job_data in self._airflow_jobs.items()
                if self._validate_airflow_job(job_id, job_data).is_left}

    def _deploy_rename_jobs(self):
        futures = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            for job_info in self._rename_jobs():
                futures.append(executor.submit(lambda j_info=job_info: self._delete_job_for_renamed_job(j_info)))

        return await_futures_and_update_states(futures, self._operation_to_step_id[Operation.Rename])

    def _delete_job_for_renamed_job(self, jobs_info: JobInfo):
        try:
            client = create_airflow_client(jobs_info.fabric_id, self._project_config)
            client.delete_dag(sanitize_job(jobs_info.external_job_id))

            log(f"{Colors.OKGREEN}Successfully deleted dag for job_id: {jobs_info.id}{Colors.ENDC}",
                step_id=self._RENAME_JOBS_STEP_ID)

            return Either(right=JobInfoAndOperation(jobs_info, OperationType.DELETED))
        except Exception as e:
            log(f"{Colors.WARNING}Failed to delete dag for job_id: {jobs_info.id} with job name {jobs_info.external_job_id} in fabric {jobs_info.fabric_id}, Please delete the Dag manually from Dag location if required.{Colors.ENDC}",
                exception=e, step_id=self._RENAME_JOBS_STEP_ID)
            return Either(right=JobInfoAndOperation(jobs_info, OperationType.DELETED))

    def _deploy_pause_jobs(self):
        futures = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            for job_id, job_data in self._pause_jobs().items():
                futures.append(executor.submit(lambda j_id=job_id, j_data=job_data: self._pause_job(j_id, j_data)))

        return await_futures_and_update_states(futures, self._operation_to_step_id[Operation.Pause])

    def _pause_job(self, job_id: str, job_info: JobInfo):
        client = self.get_airflow_client(job_info.fabric_id)

        if len(job_info.external_job_id) > 0:
            dag_name = job_info.external_job_id
        else:
            dag_name = sanitize_job(job_id)

        job_info.pause(True)
        try:
            client.pause_dag(dag_name)
            log(f"{Colors.OKGREEN}Successfully paused job {dag_name} for job_id {job_id}{Colors.ENDC}",
                step_id=self._PAUSE_JOBS_STEP_ID)
            return Either(right=JobInfoAndOperation(job_info, OperationType.REFRESH))
        except Exception as e:
            log(f"{Colors.WARNING}Failed to pause_dag for job_id: {job_id} with dag_name {dag_name}{Colors.ENDC}",
                exception=e,
                step_id=self._PAUSE_JOBS_STEP_ID)
            return Either(right=JobInfoAndOperation(job_info, OperationType.REFRESH))


class AirflowGitSecrets:
    _AIRFLOW_GIT_SECRETS_STEP_ID = "airflow-git-secrets"

    def __init__(self, project: Project, airflow_jobs: AirflowJobDeployment,
                 project_config: ProjectConfig):
        self.project = project
        self.airflow_jobs = airflow_jobs
        self.project_config = project_config
        self.deployment_state = project_config.jobs_state
        self.fabric_config = project_config.fabric_config

    def summary(self) -> List[str]:
        summary = []

        if len(self.airflow_jobs.prophecy_managed_dbt_jobs) > 0:

            for project_git_tokens in self.fabric_config.project_git_tokens:
                git_tokens = project_git_tokens.git_token
                project_id = project_git_tokens.project_id

                if len(git_tokens) > 0:
                    summary.append(f"Creating git secrets for project {project_id} ")

        return summary

    def headers(self) -> List[StepMetadata]:
        if len(self.airflow_jobs.prophecy_managed_dbt_jobs) > 0:
            return [
                StepMetadata(self._AIRFLOW_GIT_SECRETS_STEP_ID, "Create git secrets for airflow jobs", Operation.Build,
                             StepType.AirflowGitSecrets)]
        else:
            return []

    def deploy(self):

        futures = []

        if len(self.airflow_jobs.prophecy_managed_dbt_jobs) > 0:

            if len(self.fabric_config.project_git_tokens) > 0:
                log(f"\n\n{Colors.OKBLUE} Uploading project git tokens {Colors.ENDC}\n\n")

            job_data = list(self.airflow_jobs.prophecy_managed_dbt_jobs.values())[0]

            with ThreadPoolExecutor(max_workers=10) as executor:
                for project_git_tokens in self.fabric_config.project_git_tokens:
                    git_tokens = project_git_tokens.git_token
                    project_id = project_git_tokens.project_id

                    if len(git_tokens) == 0:
                        log(f"{Colors.WARNING}No git tokens found for project_id: {project_id}, Ignoring{Colors.ENDC}",
                            step_id=self._AIRFLOW_GIT_SECRETS_STEP_ID)

                    else:
                        futures.append(
                            executor.submit(
                                lambda p_id=project_id, j_data=job_data, token=git_tokens: self._create_git_secrets(
                                    p_id, j_data, token)))

            return await_futures_and_update_states(futures, self._AIRFLOW_GIT_SECRETS_STEP_ID)
        else:
            # log("No dbt jobs found, Ignoring", step_id=self._AIRFLOW_GIT_SECRETS_STEP_ID)
            return []

    def _create_git_secrets(self, project_id, job_data, git_tokens):
        # todo fix this
        execution_db_suffix = os.getenv('EXECUTION_DB_SUFFIX', 'dev')
        client = self.airflow_jobs.get_airflow_client(job_data.fabric_id)
        try:
            client.create_secret(
                generate_secure_content(f'{execution_db_suffix}_{project_id}', 'gitSecretSalt'), git_tokens)
            log(f'{Colors.OKGREEN}Successfully created git secrets for project {project_id}{Colors.ENDC}',
                step_id=self._AIRFLOW_GIT_SECRETS_STEP_ID)

        except Exception as e:
            log(f'{Colors.WARNING}Failed in creating git secrets for project {project_id}{Colors.ENDC}', exception=e,
                step_id=self._AIRFLOW_GIT_SECRETS_STEP_ID)


class EMRPipelineConfigurations:
    _STEP_ID = "EMR_pipeline_configurations"

    def __init__(self, project: Project, airflow_jobs: AirflowJobDeployment,
                 project_config: ProjectConfig):
        self.airflow_jobs = airflow_jobs
        self.pipeline_configurations = project.pipeline_configurations
        self.project_config = project_config
        self.project = project
        self._jobs_state = project_config.jobs_state
        self._fabric_config = project_config.fabric_config
        self._rest_client_factory = RestClientFactory.get_instance(RestClientFactory, self._fabric_config)

    def _emr_fabrics(self):
        return self._fabric_config.emr_fabrics()

    def summary(self) -> List[str]:
        summary = []
        for ids in self.pipeline_configurations.keys():
            if len(self._emr_fabrics()) > 0:
                summary.append(f"Uploading pipeline emr-configurations for pipeline {ids}")

        return summary

    def headers(self) -> List[StepMetadata]:
        all_configs = [value for sublist in self.pipeline_configurations.values() for value in sublist]
        if len(all_configs) > 0 and len(self._emr_fabrics()) > 0:
            return [StepMetadata(self._STEP_ID,
                                 f"Upload {len(all_configs)} pipeline emr-configurations",
                                 Operation.Upload, StepType.PipelineConfiguration)]
        else:
            return []

    def deploy(self):

        futures = []

        with ThreadPoolExecutor(max_workers=10) as executor:
            if len(self.headers()) > 0:
                log(f"\n\n{Colors.OKBLUE}Uploading EMR pipeline configurations {Colors.ENDC}\n\n")

            def execute_job(_fabric_info, _config_content, _config_path):
                futures.append(executor.submit(
                    lambda f_info=_fabric_info, conf_content=_config_content, conf_path=_config_path:
                    self._upload_configuration(f_info, conf_content, conf_path)))

            for pipeline_id, configurations in self.pipeline_configurations.items():

                path = self.project_config.system_config.get_s3_base_path()
                pipeline_path = f'{path}/{self.project.project_id}/{self.project.release_version}/configurations/{pipeline_id}'

                for configuration_name, configuration_content in configurations.items():
                    configuration_path = f'{pipeline_path}/{configuration_name}.jsn'

                    for fabric_info in self._fabric_config.emr_fabrics():

                        if fabric_info.emr is not None:
                            execute_job(fabric_info, configuration_content, configuration_path)

        return await_futures_and_update_states(futures, self._STEP_ID)

    def _upload_configuration(self, fabric_info: FabricInfo, configuration_content, configuration_path):
        upload_path = f"{fabric_info.emr.bare_path_prefix()}/{configuration_path}"
        emr_info = fabric_info.emr

        try:
            client = self._rest_client_factory.s3_client(str(fabric_info.id))
            client.upload_content(emr_info.bare_bucket(), upload_path, configuration_content)

            log(f"{Colors.OKGREEN}Uploaded pipeline configuration on path {upload_path} on fabric {fabric_info.id}{Colors.ENDC}",
                step_id=self._STEP_ID)

            return Either(right=True)

        except Exception as e:
            log(f"{Colors.FAIL}Failed to upload pipeline configuration for path {upload_path} for fabric {fabric_info.id}{Colors.ENDC}",
                exception=e,
                step_id=self._STEP_ID)
            return Either(left=e)


class DataprocPipelineConfigurations:
    _STEP_ID = "DATAPROC_pipeline_configurations"

    def __init__(self, project: Project, airflow_jobs: AirflowJobDeployment,
                 project_config: ProjectConfig):
        self.airflow_jobs = airflow_jobs
        self.pipeline_configurations = project.pipeline_configurations
        self.project = project
        self.project_config = project_config
        self._deployment_state = project_config.jobs_state
        self._fabric_config = project_config.fabric_config
        self._rest_client_factory = RestClientFactory.get_instance(RestClientFactory, self._fabric_config)

    def _dataproc_fabrics(self):
        return self._fabric_config.dataproc_fabrics()

    def summary(self) -> List[str]:
        summary = []
        for ids in self.pipeline_configurations.keys():
            if len(self._dataproc_fabrics()) > 0:
                summary.append(f"Uploading pipeline dataproc-configurations for pipeline {ids}")

        return summary

    def headers(self) -> List[StepMetadata]:
        all_configs = [value for sublist in self.pipeline_configurations.values() for value in sublist]
        if len(all_configs) > 0 and len(self._dataproc_fabrics()) > 0:
            return [StepMetadata(self._STEP_ID,
                                 f"Upload {len(all_configs)} pipeline dataproc-configurations",
                                 Operation.Upload, StepType.PipelineConfiguration)]
        else:
            return []

    def deploy(self):
        futures = []

        with ThreadPoolExecutor(max_workers=10) as executor:
            if len(self.headers()) > 0:
                log(f"\n\n{Colors.OKBLUE} Uploading dataproc configurations {Colors.ENDC}\n\n")

            def execute_job(fabric_info, configuration_content, configuration_path):
                futures.append(executor.submit(
                    lambda f_info=fabric_info, conf_content=configuration_content, conf_path=configuration_path:
                    self._upload_configuration(f_info, conf_content, conf_path)))

            for pipeline_id, configurations in self.pipeline_configurations.items():
                path = self.project_config.system_config.get_s3_base_path()
                pipeline_path = f'{path}/{self.project.project_id}/{self.project.release_version}/configurations/{pipeline_id}'
                for configuration_name, configuration_content in configurations.items():
                    configuration_path = f'{pipeline_path}/{configuration_name}.jsn'

                    for fabric_info in self._fabric_config.dataproc_fabrics():

                        if fabric_info.dataproc is not None:
                            execute_job(fabric_info, configuration_content, configuration_path)

        return await_futures_and_update_states(futures, self._STEP_ID)

    def _upload_configuration(self, fabric_info: FabricInfo, configuration_content, configuration_path):
        upload_path = f"{fabric_info.dataproc.bare_path_prefix()}/{configuration_path}"
        dataproc_info = fabric_info.dataproc
        try:
            client = self._rest_client_factory.dataproc_client(str(fabric_info.id))
            client.put_object(dataproc_info.bare_bucket(), upload_path, configuration_content)

            log(f"{Colors.OKGREEN}Uploaded pipeline configuration on path {upload_path}{Colors.ENDC}",
                step_id=self._STEP_ID)

            return Either(right=True)

        except Exception as e:
            log(f"{Colors.FAIL}Failed to upload pipeline configuration for path {upload_path} for fabric {fabric_info.id}{Colors.ENDC}",
                exception=e,
                step_id=self._STEP_ID)
            return Either(left=e)



from .composer import ComposerRestClient
from .mwaa import MWAARestClient
from ...utils.exceptions import UnknownAirflowProviderException, FabricNotConfiguredException
from ...utils.project_config import ProjectConfig, FabricType, FabricProviderType


def create_airflow_client(fabric_id: str, project_config: ProjectConfig):
    fabric_info = project_config.fabric_config.get_fabric(fabric_id)

    if fabric_info is not None and fabric_info.type == FabricType.Airflow:

        composer = fabric_info.composer
        mwaa = fabric_info.mwaa

        if composer is not None:

            return ComposerRestClient(composer.airflow_url, composer.project_id, composer.client_id,
                                      composer.key_json, composer.dag_location)

        elif mwaa is not None:
            return MWAARestClient(mwaa.environment_name, mwaa.region, mwaa.access_key,
                                  mwaa.secret_key)

        else:
            raise UnknownAirflowProviderException("unknown provider")

    else:
        raise FabricNotConfiguredException(f"Fabric {fabric_id} is not configured in state config")


def get_fabric_provider_type(fabric_id: str, project_config: ProjectConfig) -> str:
    fabric_info = project_config.fabric_config.get_fabric(fabric_id)
    if fabric_info is not None:
        return str(fabric_info.provider.value)
    else:
        return FabricProviderType.Databricks.value



class PBTException(ValueError):
    def __init__(self, message):
        self.message = message


class DagNotAvailableException(PBTException):
    def __init__(self, message):
        self.message = message


class DagFileDeletionFailedException(PBTException):
    def __init__(self, message, e):
        self.message = message
        self.e = e


class DagUploadFailedException(PBTException):
    def __init__(self, message, e):
        self.message = message
        self.e = e


class DagListParsingFailedException(PBTException):

    def __init__(self, message, e):
        self.message = message
        self.e = e


class ProjectPathNotFoundException(PBTException):
    def __init__(self, message):
        self.message = message


class ProjectFileNotFoundException(PBTException):
    def __init__(self, message):
        self.message = message


class UnknownAirflowProviderException(PBTException):
    def __init__(self, message):
        self.message = message


class FabricNotConfiguredException(PBTException):
    def __init__(self, message):
        self.message = message


class ArtifactDownloadFailedException(PBTException):
    def __init__(self, response):
        self.message = f"Artifact download failed with status code {response.status_code} and reason {response.text}"


class InvalidFabricException(PBTException):
    def __init__(self, message):
        self.message = message


class ProjectBuildFailedException(PBTException):
    def __init__(self, message):
        self.message = message


class PipelineBuildFailedException(PBTException):
    def __init__(self, message):
        self.message = message


class ConfigFileNotFoundException(PBTException):
    def __init__(self, message):
        self.message = message


class PipelinePathNotFoundException(PBTException):
    def __init__(self, message):
        self.message = message


class DuplicateJobNameException(PBTException):
    def __init__(self, message):
        self.message = message



import enum
import os
from asyncio import Future
from concurrent.futures import as_completed
from typing import List, Optional

from pydantic import BaseModel
from pydantic_yaml import parse_yaml_raw_as

from .constants import PROPHECY_ARTIFACTS, DBFS_FILE_STORE
from .exceptions import ConfigFileNotFoundException
from .project_models import Status
from ..deployment import OperationType, JobInfoAndOperation
from ..entities.project import Project
from ..utility import Either, custom_print as log, is_online_mode


class SchedulerType(enum.Enum):
    Composer = "Composer"
    MWAA = "MWAA"
    Databricks = "Databricks"
    Prophecy = "Prophecy"
    EMR = "EMR"

    @staticmethod
    def from_type(provider_type: str):
        return SchedulerType[provider_type]


class FabricType(enum.Enum):
    Spark = "Spark"
    Sql = "Sql"
    Airflow = "Airflow"


class FabricProviderType(enum.Enum):
    Composer = "Composer"
    MWAA = "MWAA"
    Databricks = "Databricks"
    Prophecy = "Prophecy"
    EMR = "EMR"
    Dataproc = "Dataproc"


class DeploymentMode(enum.Enum):
    FullProject = "FullProject"
    SelectiveJob = "SelectiveJob"


class BucketAndPathExtractor:
    def __init__(self, bucket: str, prefix: str):
        self.bucket = bucket
        self.prefix = prefix

    def path(self):
        return self.bucket.replace(self.prefix, "").split("/")

    def bare_bucket(self):
        return self.path()[0]

    def bare_path_prefix(self):
        return "/".join(self.path()[1:])


#
class EMRInfo(BaseModel):
    region: str
    bucket: str
    access_key_id: str
    secret_access_key: str
    session_token: Optional[str] = None

    def bucket_path_extractor(self):
        return BucketAndPathExtractor(self.bucket, "s3://")

    def bare_bucket(self):
        return self.bucket_path_extractor().bare_bucket()

    def bare_path_prefix(self):
        return self.bucket_path_extractor().bare_path_prefix()


class DataprocInfo(BaseModel):
    bucket: str
    project_id: str
    key_json: str
    location: str

    def bucket_path_extractor(self):
        return BucketAndPathExtractor(self.bucket, "gs://")

    def bare_bucket(self):
        return self.bucket_path_extractor().bare_bucket()

    def bare_path_prefix(self):
        return self.bucket_path_extractor().bare_path_prefix()


class ComposerInfo(BaseModel):
    key_json: str
    version: str
    location: Optional[str] = None
    project_id: str
    airflow_url: str
    dag_location: str
    client_id: Optional[str] = None


class MwaaInfo(BaseModel):
    region: str
    version: str
    access_key: str
    secret_key: str
    airflow_url: str
    dag_location: str
    environment_name: str


class DatabricksInfo(BaseModel):
    url: str
    token: str

    @staticmethod
    def create(host: str, token: str):
        return DatabricksInfo(url=host, token=token)


class FabricInfo(BaseModel):
    id: str
    name: str
    type: Optional[FabricType]  # sql/ databricks/ airflow
    provider: Optional[FabricProviderType]  # composer/mwaa/prophecy/databricks/dataproc
    composer: Optional[ComposerInfo] = None
    mwaa: Optional[MwaaInfo] = None
    databricks: Optional[DatabricksInfo] = None
    emr: Optional[EMRInfo] = None
    dataproc: Optional[DataprocInfo] = None

    @staticmethod
    def create_db_fabric(id: str, host: str, token: str):
        return FabricInfo(id=id, name="", type=FabricType.Spark, provider=FabricProviderType.Databricks,
                          databricks=DatabricksInfo.create(host, token))


class JobInfo(BaseModel):
    id: str
    name: str
    type: SchedulerType
    external_job_id: str
    fabric_id: str
    is_paused: Optional[bool] = False
    skip_processing: Optional[bool] = False  # this is useful in case when we deploy from older release tags.
    release_tag: Optional[str] = None

    def update_release_tag(self, release_tag):
        self.release_tag = release_tag

    def is_job_same_as(self, job_info) -> bool:
        return self.external_job_id == job_info.external_job_id and self.fabric_id == job_info.fabric_id and \
            self.id == job_info.id and self.name == job_info.name and self.type == job_info.type

    def pause(self, flag: bool):
        self.is_paused = flag

    @staticmethod
    def create_job(name: str, id: str, fabric_id: str, external_job_id: str, release_tag: str,
                   is_paused: bool = False, fabric_provider_type: str = "Databricks"):
        return JobInfo(name=name,
                       type=SchedulerType.from_type(fabric_provider_type),
                       id=id,
                       fabric_id=fabric_id,
                       external_job_id=external_job_id,
                       release_tag=release_tag,
                       is_paused=is_paused)


class ProjectAndGitTokens(BaseModel):
    project_id: str
    git_token: str = ""


class FabricConfig(BaseModel):
    fabrics: List[FabricInfo] = []
    project_git_tokens: List[ProjectAndGitTokens] = []

    def get_fabric(self, fabric_id: str) -> Optional[FabricInfo]:
        return next((fabric for fabric in self.fabrics if fabric.id == fabric_id), None)

    def does_fabric_exist(self, fabric_id: str) -> bool:
        return self.get_fabric(fabric_id) is not None

    def git_token_for_project(self, project_id: str) -> Optional[str]:
        return next((entity.git_token for entity in self.project_git_tokens if entity.project_id == project_id),
                    None)

    def is_prophecy_managed(self, fabric_id: str) -> bool:
        if self.get_fabric(fabric_id) is not None:
            fabric_info = self.get_fabric(fabric_id)
            return fabric_info.type == FabricType.Airflow and fabric_info.provider == FabricProviderType.Prophecy

        return False

    def is_fabric_emr_fabric(self, fabric_id: str) -> bool:
        return any((fabric for fabric in self.fabrics if
                    fabric.id == fabric_id and fabric.provider == FabricProviderType.EMR))

    def db_fabrics(self) -> List[str]:
        return [fabric.id for fabric in self.fabrics if fabric.provider == FabricProviderType.Databricks]

    def is_databricks_fabric(self, fabric_id: str) -> bool:
        return fabric_id in self.db_fabrics()

    def emr_fabrics(self) -> List[FabricInfo]:
        return [fabric for fabric in self.fabrics if
                (fabric.type == FabricType.Spark and fabric.provider == FabricProviderType.EMR)]

    def dataproc_fabrics(self) -> List[FabricInfo]:
        return [fabric for fabric in self.fabrics if
                (fabric.type == FabricType.Spark and fabric.provider == FabricProviderType.Dataproc)]

    def list_all_fabrics(self) -> List[str]:
        return [fabric.id for fabric in self.fabrics]

    def filter_provided_fabrics(self, fabric_ids: str):
        if fabric_ids is not None and len(fabric_ids) > 0:
            fabrics = set(fabric_ids.split(","))

            self.fabrics = [fabric for fabric in self.fabrics if fabric.id in fabrics]


class JobsState(BaseModel):
    version: str = "1.0"
    jobs: List[JobInfo] = []

    @staticmethod
    def empty():
        state = JobsState(version="1.0")
        return state

    @property
    def jobs_to_process(self) -> List[JobInfo]:
        return [job for job in self.jobs if not job.skip_processing]

    @property
    def databricks_jobs(self) -> List[JobInfo]:
        return [job for job in self.jobs_to_process if job.type == SchedulerType.Databricks]

    @property
    def airflow_jobs(self) -> List[JobInfo]:
        return [job for job in self.jobs_to_process if job.type != SchedulerType.Databricks]

    def contains_job(self, job_id: str, fabric_uid: str) -> bool:
        return any(
            job.id == job_id and job.fabric_id == fabric_uid for job in self.jobs_to_process)

    def all_jobs_for_id(self, job_id: str) -> List[JobInfo]:
        return [job for job in self.jobs_to_process if job.id == job_id]

    def job_for_id_and_fabric(self, job_id: str, fabric_id: str) -> Optional[JobInfo]:
        return next(
            (job for job in self.all_jobs_for_id(job_id) if job.fabric_id == fabric_id), None)

    def filter_job(self, job_info) -> List[JobInfo]:
        return [job for job in self.jobs if
                not (job.id == job_info.id and job.fabric_id == job_info.fabric_id)]

    def update_state(self, jobs_responses: List[Either]):
        filtered_response: List[JobInfoAndOperation] = [response.right for
                                                        response
                                                        in jobs_responses if
                                                        response.is_right]
        # Important to do all operations in this order,
        # first we delete
        # then we refresh
        # then we create

        for job_response in filtered_response:
            job_info = job_response.job_info

            if job_response.operation == OperationType.DELETED:
                self.jobs = self.filter_job(job_info)

        for job_response in filtered_response:
            job_info = job_response.job_info

            if job_response.operation == OperationType.REFRESH:
                new_jobs = self.filter_job(job_info)
                new_jobs.append(job_info)
                self.jobs = new_jobs

        for job_response in filtered_response:
            job_info = job_response.job_info

            if job_response.operation == OperationType.CREATED:
                matching_job = next((job for job in self.jobs if job.id == job_info.id), None)

                new_jobs = self.filter_job(job_info)

                if matching_job is not None:
                    # super important to preserve the release tag.
                    job_info.update_release_tag(matching_job.release_tag)

                new_jobs.append(job_info)

                self.jobs = new_jobs

    def filter_provided_jobs(self, job_ids):
        if job_ids is not None and len(job_ids) > 0:
            jobs = set([f"jobs/{job}" for job in job_ids.split(",")])

            self.jobs = [job for job in self.jobs if job.id in jobs]


class JobAndFabric(BaseModel):
    job_id: str
    fabric_id: str

    @staticmethod
    def create(job: str, fabric: str):
        return JobAndFabric(job_id=job, fabric_id=fabric)


class ConfigsOverride(BaseModel):
    tests_enabled: Optional[bool] = False
    mode: Optional[DeploymentMode] = DeploymentMode.FullProject
    jobs_and_fabric: Optional[List[JobAndFabric]] = None

    def find_fabric_override_for_job(self, job_id: str) -> Optional[str]:
        if self.mode == DeploymentMode.SelectiveJob:
            return next(
                (entity.fabric_id for entity in self.jobs_and_fabric if entity.job_id == job_id),
                None
            )
        return None

    def is_job_to_run(self, job_id) -> bool:
        if self.mode == DeploymentMode.SelectiveJob:
            return any(job_and_fabric.job_id == job_id for job_and_fabric in self.jobs_and_fabric)
        return True

    @staticmethod
    def empty():
        return ConfigsOverride()

    def filter_jobs(self, job_ids):
        if job_ids is not None and len(job_ids) > 0:
            if self.jobs_and_fabric is not None:
                filtered_jobs = set([f"jobs/{job}" for job in job_ids.split(",")])

                self.jobs_and_fabric = [jobs for jobs in self.jobs_and_fabric if jobs.job_id in filtered_jobs]
                self.mode = DeploymentMode.SelectiveJob
            else:
                self.jobs_and_fabric = [JobAndFabric.create(f"jobs/{job}", "") for job in job_ids.split(",")]
                self.mode = DeploymentMode.SelectiveJob


class NexusConfig(BaseModel):
    url: str
    username: str
    password: str
    repository: str


class SystemConfig(BaseModel):
    customer_name: Optional[str] = 'dev'
    control_plane_name: Optional[str] = 'execution'
    prophecy_salt: Optional[str] = 'execution'
    nexus: Optional[NexusConfig] = None

    def get_dbfs_base_path(self):
        return f'{DBFS_FILE_STORE}/{PROPHECY_ARTIFACTS}/{self.customer_name}/{self.control_plane_name}'

    def get_s3_base_path(self):
        return f"{PROPHECY_ARTIFACTS}/{self.customer_name}/{self.control_plane_name}"

    @staticmethod
    def empty():
        return SystemConfig()


def load_jobs_state(job_state_path: str, is_based_on_file: bool = True):
    if job_state_path is not None and len(job_state_path) > 0 and is_based_on_file:
        with open(job_state_path, "r") as job_state:
            return parse_yaml_raw_as(JobsState, job_state.read())
    else:
        raise ConfigFileNotFoundException("Job state config path is not provided")


def load_system_config(system_config_path: str, is_based_on_file: bool = True):
    if system_config_path is not None and len(system_config_path) > 0 and is_based_on_file:
        with open(system_config_path, "r") as system_config:
            return parse_yaml_raw_as(SystemConfig, system_config.read())
    else:
        raise ConfigFileNotFoundException("System config path is not provided")


def load_configs_override(configs_override_path, is_based_on_file: bool = True):
    if configs_override_path is not None and len(configs_override_path) > 0 and is_based_on_file:
        with open(configs_override_path, "r") as config_override:
            return parse_yaml_raw_as(ConfigsOverride, config_override.read())
    else:
        return ConfigsOverride.empty()


def load_fabric_config(fabric_config_path):
    if fabric_config_path is not None and len(fabric_config_path) > 0:
        with open(fabric_config_path, "r") as fabric_config:
            return parse_yaml_raw_as(FabricConfig, fabric_config.read())
    else:
        raise ConfigFileNotFoundException("Fabric config path is not provided")


class ProjectConfig:
    def __init__(self, jobs_state: JobsState, fabric_config: FabricConfig, system_config: SystemConfig,
                 config_override: ConfigsOverride, based_on_file: bool = True, skip_builds: bool = False,
                 migrate: bool = False, conf_folder: str = ""):
        self.jobs_state = jobs_state
        self.fabric_config = fabric_config
        self.system_config = system_config
        self.configs_override = config_override
        self.based_on_file = based_on_file
        self.skip_builds = skip_builds
        self.migrate = migrate
        self.conf_folder = conf_folder

    @staticmethod
    def from_path(project: Project, job_state_path: str, system_config_path: str, configs_override_path: str,
                  fabric_config_path: str, fabric_ids: str, job_ids: str, skip_build: bool, conf_folder: str,
                  migrate: bool):

        is_based_on_file = conf_folder != "" and len(conf_folder) > 0

        if is_online_mode() and len(conf_folder) > 0:
            jobs = load_jobs_state(job_state_path)
            fabrics = load_fabric_config(fabric_config_path)
            system = load_system_config(system_config_path)
            configs = load_configs_override(configs_override_path)
            return ProjectConfig(jobs, fabrics, system, configs, skip_builds=skip_build)

        else:

            if not is_based_on_file:
                # only cli case for databricks/ fabrics
                host = os.environ.get("DATABRICKS_HOST", 'test')
                token = os.environ.get("DATABRICKS_TOKEN", 'test')
                if not fabric_ids:
                    allowed_fabrics = project.fabrics()
                else:
                    allowed_fabrics = fabric_ids.split(",")
                fabric_list = [FabricInfo.create_db_fabric(id=fabric, host=host, token=token) for fabric in
                               allowed_fabrics]
                fabrics_config = FabricConfig(fabrics=fabric_list)

            else:
                # most important files.
                # fabrics
                fabrics_config = load_fabric_config(fabric_config_path)
                fabrics_config.filter_provided_fabrics(fabric_ids)

            # jobs
            try:
                jobs = load_jobs_state(job_state_path, is_based_on_file)
            except ConfigFileNotFoundException:
                jobs = JobsState.empty()

                # system
            try:
                system = load_system_config(system_config_path, is_based_on_file)
            except ConfigFileNotFoundException:
                # system
                system = SystemConfig.empty()
                system.customer_name = project.find_customer_name()
                system.control_plane_name = project.find_control_plane_name()

            # configs
            try:
                configs = load_configs_override(configs_override_path, is_based_on_file)
                configs.filter_jobs(job_ids)
                jobs.filter_provided_jobs(job_ids)
            except ConfigFileNotFoundException:
                configs = ConfigsOverride.empty()
                configs.mode = DeploymentMode.FullProject  # only build what's needed.
                if len(job_ids) != 0:
                    allowed_jobs = job_ids.split(",")
                    configs.jobs_and_fabric = [JobAndFabric.create(f"jobs/{job}", "") for job in allowed_jobs]
                    configs.mode = DeploymentMode.SelectiveJob

            return ProjectConfig(jobs, fabrics_config, system, configs, based_on_file=is_based_on_file,
                                 skip_builds=skip_build, migrate=migrate)

    # best used when invoking from execution.
    @classmethod
    def from_conf_folder(cls, project: Project, conf_folder, fabric_ids: str, job_ids: str, skip_builds: bool,
                         migrate: bool):
        jobs_state = os.path.join(conf_folder, "state.yml")
        system_config = os.path.join(conf_folder, "system.yml")
        config_override = os.path.join(conf_folder, "override.yml")
        fabric_config = os.path.join(conf_folder, "fabrics.yml")

        return ProjectConfig.from_path(project, jobs_state, system_config, config_override, fabric_config, fabric_ids,
                                       job_ids,
                                       skip_builds, conf_folder, migrate)


def await_futures_and_update_states(futures: List[Future], step_id: str):
    responses = []

    for future in as_completed(futures):
        responses.append(future.result())

    update_state(responses, step_id)
    return responses


def update_state(responses: List, step_id: str):
    if responses is not None and len(responses) > 0:
        if any(response.is_left for response in responses):
            log(step_status=Status.FAILED, step_id=step_id)
        else:
            log(step_status=Status.SUCCEEDED, step_id=step_id)



import enum
import json
import traceback
from typing import List, Optional

from pydantic import BaseModel


class Operation(enum.Enum):
    Build = "Build"
    Add = "Add"
    Delete = "Delete"
    Remove = "Remove"
    Refresh = "Refresh"
    Skipped = "Skipped"
    Pause = "Pause"
    Rename = "Rename"
    Upload = "Upload"


class StepType(enum.Enum):
    Pipeline = "Pipeline"
    PipelineConfiguration = "PipelineConfiguration"
    Job = "Job"
    Script = "Script"
    DbtProfile = "DbtProfile"
    DbtSecret = "DbtSecret"
    AirflowGitSecrets = "AirflowGitSecrets"
    Summary = "Summary"
    Subgraph = "Subgraph"  # why this?
    Project = "Project"
    Gems = "Gems"


class LogLevel(enum.Enum):
    INFO = "INFO"
    DEBUG = "DEBUG"
    WARN = "WARN"
    ERROR = "ERROR"
    SUCCESS = "SUCCESS"
    TRACE = "TRACE"


class Colors:
    HEADER = '\033[95m'  # Purple
    OKBLUE = '\033[94m'  # Blue
    MAGENTA = '\033[35m'
    OKCYAN = '\033[96m'  # Cyan
    OKGREEN = '\033[92m'  # Green
    WARNING = '\033[93m'  # Yellow
    FAIL = '\033[91m'  # Red
    ENDC = '\033[0m'  # Reset color


def to_dict_recursive(obj):
    if isinstance(obj, enum.Enum):
        return obj.name

    if isinstance(obj, (list, tuple)):
        return [to_dict_recursive(item) for item in obj]
    elif isinstance(obj, dict):
        # Skip keys with None values
        return {key: to_dict_recursive(value) for key, value in obj.items() if value is not None}
    elif hasattr(obj, '__dict__'):
        return to_dict_recursive(
            {key: value for key, value in obj.__dict__.items() if not isinstance(value, type) and value is not None})
    elif hasattr(obj, '__slots__'):
        return to_dict_recursive(
            {slot: getattr(obj, slot) for slot in obj.__slots__ if getattr(obj, slot) is not None})
    else:
        return obj


class StepMetadata:

    def __init__(self, id: str, heading: str, operation: Operation, type: StepType):
        self.id = id.replace("\\W", "_")
        self.heading = heading
        self.operation = operation
        self.type = type

    def to_json(self):
        return json.dumps(self, default=to_dict_recursive)


class LogType(enum.Enum):
    Header = "Header"
    LogLine = "LogLine"
    Status = "Status"


class Status(enum.Enum):
    RUNNING = "RUNNING"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"


class StepStatus:
    def __init__(self, step_id: str, status: Status):
        self.step_id = step_id
        self.status = status

    def to_json(self):
        # we want json to be in single line
        return json.dumps(self, default=to_dict_recursive)


class LogLine:
    def __init__(self, uid: str, log: str, level: LogLevel.INFO):
        self.uid = uid
        self.log = log
        self.level = level

    def to_json(self):
        # we want json to be in single line
        return json.dumps(self, default=to_dict_recursive)


class LogEvent:
    def __init__(self, step_id: str, type: LogType, log: Optional[str] = None):
        self.step_id = step_id
        self.type = type
        self.log = log

    def to_json(self):
        # we want json to be in single line
        return json.dumps(self, default=to_dict_recursive)

    @staticmethod
    def from_step_metadata(step_metadata: StepMetadata):
        return LogEvent(step_metadata.id, LogType.Header, step_metadata.to_json())

    @staticmethod
    def from_log(step_id: str, log: str, ex: Optional[Exception] = None, level: Optional[LogLevel] = None):
        log_level = level if level is not None else LogLevel.INFO
        if ex is not None:
            captured_trace = traceback.format_exc(limit=200)
            log = f"{log} exception message: {str(ex)}: Stacktrace: {'  '.join(captured_trace.splitlines())}"

            if log_level is not None:
                log_line = LogLine(step_id, log, level=log_level).to_json()
            else:
                log_line = LogLine(step_id, log, level=LogLevel.ERROR).to_json()

        else:
            log_line = LogLine(step_id, log, level=log_level).to_json()

        return LogEvent(step_id, LogType.LogLine, log_line)

    @staticmethod
    def from_status(status: Status, step_id: str):
        return LogEvent(step_id, LogType.Status, StepStatus(step_id, status).to_json())


class Component(BaseModel):
    id: str
    node_name: str
    path: Optional[str]
    language: str


class JobContent(BaseModel):
    job_id: str
    components: List[Component]
    enabled: bool
    fabric_id: str
    secret_scope: Optional[str]


class DbtComponentsModel:
    def __init__(self, fabric_id: str, secret_scope: str, components: List):
        self.fabric_id = str(fabric_id)
        self.secret_scope = secret_scope
        self.components = components


class ScriptComponentsModel:
    def __init__(self, fabric_id: str, scripts: List):
        self.scripts = scripts
        self.fabric_id = fabric_id


class DAG:

    def __init__(self, dag_id: str, description: Optional[str] = None, file_token: Optional[str] = None,
                 fileloc: Optional[str] = None,
                 is_active: Optional[bool] = None, is_paused: bool = True, is_subdag: Optional[bool] = None,
                 owners: List[str] = None,
                 root_dag_id: Optional[str] = None, schedule_interval: Optional[str] = None,
                 next_dagrun: Optional[str] = None,
                 tags: List[str] = []):
        self.dag_id = dag_id
        self.description = description
        self.file_token = file_token
        self.fileloc = fileloc
        self.is_active = is_active
        self.is_paused = is_paused
        self.is_subdag = is_subdag
        self.owners = owners
        self.root_dag_id = root_dag_id
        self.schedule_interval = schedule_interval
        self.next_dagrun = next_dagrun
        self.tags = tags

    @staticmethod
    def create(responses: dict):
        dag_id = responses.get('dag_id')
        description = responses.get('description', None)
        file_token = responses.get('file_token', None)
        fileloc = responses.get('fileloc', None)
        is_active = responses.get('is_active', None)
        is_paused = responses.get('is_paused', None)
        is_subdag = responses.get('is_subdag', None)
        owners = responses.get('owners', [])
        root_dag_id = responses.get('root_dag_id', None)

        try:
            schedule_interval = responses.get('schedule_interval', {}).get('value', None)
        except AttributeError:
            schedule_interval = None

        next_dagrun = responses.get('next_dagrun', None)
        tags = responses.get('tags', [])
        return DAG(dag_id, description, file_token, fileloc, is_active, is_paused, is_subdag, owners, root_dag_id,
                   schedule_interval, next_dagrun, tags)

    # different from the scala release.
    @staticmethod
    def create_from_mwaa(response: dict):
        dag_id = response.get('dag_id')
        fileloc = response.get('filepath', None)
        is_paused = response.get('paused', None)
        owner = response.get('owner', None)
        owners = [] if owner is None else ([owner] if isinstance(owner, str) else owner)
        return DAG(dag_id, fileloc=fileloc, is_paused=is_paused is None or bool(is_paused), owners=owners)



from click.testing import CliRunner
from src.pbt import deploy
import os

PROJECT_PATH = str(os.getcwd()) + "/test/resources/HelloWorld"
PROJECT_PATH_NEW = str(os.getcwd()) + "/test/resources/ProjectCreatedOn160523"
if os.environ.get("DATABRICKS_HOST") is None:
    os.environ["DATABRICKS_HOST"] = "test"
if os.environ.get("DATABRICKS_TOKEN") is None:
    os.environ["DATABRICKS_TOKEN"] = "test"


def test_deploy_path_default():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH])
    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert (
        "Found 4 pipelines: customers_orders1243 (python), report_top_customers (python),\njoin_agg_sort (python), "
        "farmers-markets-irs (python)" in result.output
    )
    assert "Deploying 2 jobs" in result.output
    assert "Deploying jobs for all Fabrics" in result.output
    assert "[START]:  Deploying job jobs/test-job" in result.output
    assert "[START]:  Deploying job jobs/job-another" in result.output


def test_deploy_path_default_new_project():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH_NEW, "--release-version", "1.0", "--project-id", "1"])
    assert "Found 2 jobs: AutomatedPBT-truescala, AutomatedPBTNo-truescala" in result.output
    assert "Found 2 pipelines: AutomatedPBT-truescala (scala), AutomatedPBTNo-truescala \n(scala)" in result.output
    assert "Building pipeline pipelines/AutomatedPBT-truescala" in result.output
    assert "Building pipeline pipelines/AutomatedPBTNo-truescala" in result.output
    assert "Deploying job jobs/AutomatedPBT-truescala" in result.output
    assert "Deploying job jobs/AutomatedPBTNo-truescala" in result.output

    # If running with Databricks creds on GitHub Actions
    if os.environ["DATABRICKS_HOST"] != "test":
        assert (
            "Uploading AutomatedPBT-truescala-1.0.jar to "
            "\ndbfs:/FileStore/prophecy/artifacts/prophecy/uitesting/1/1.0/pipeline/AutomatedPB\nT-truescala.jar"
            in result.output
        )
        assert (
            "Uploading AutomatedPBTNo-truescala-1.0.jar to "
            "\ndbfs:/FileStore/prophecy/artifacts/prophecy/uitesting/1/1.0/pipeline/AutomatedPB\nTNo-truescala.jar"
            in result.output
        )
        assert "[DONE]: Deployment completed successfully!" in result.output


def test_deploy_path_default_skip_builds():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH, "--skip-builds"])
    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert "[SKIP]: Skipping builds for all pipelines as '--skip-builds' flag is passed." in result.output


def test_deploy_path_fabric_id_filter():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH, "--fabric-ids", "647"])
    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert (
        "Found 4 pipelines: customers_orders1243 (python), report_top_customers (python),\njoin_agg_sort (python), "
        "farmers-markets-irs (python)" in result.output
    )
    assert "Deploying jobs only for given Fabric IDs: ['647']" in result.output
    assert "[START]:  Deploying job jobs/test-job" in result.output
    assert "[DEPLOY]: Job being deployed for fabric id: 647" in result.output
    assert "[START]:  Deploying job jobs/job-another" in result.output
    assert "[SKIP]: Job skipped as it belongs to fabric id (not passed): 648" in result.output


def test_deploy_path_pipeline_invalid_fabric_id():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH, "--fabric-ids", "999"])
    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert (
        "Found 4 pipelines: customers_orders1243 (python), report_top_customers (python),\njoin_agg_sort (python), "
        "farmers-markets-irs (python)" in result.output
    )
    assert "Deploying jobs only for given Fabric IDs: ['999']" in result.output
    assert "[START]:  Deploying job jobs/test-job" in result.output
    assert "[SKIP]: Job skipped as it belongs to fabric id (not passed): 647" in result.output
    assert "[START]:  Deploying job jobs/job-another" in result.output
    assert "[SKIP]: Job skipped as it belongs to fabric id (not passed): 648" in result.output


def test_deploy_with_fabric_id_and_job_id_filter():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH, "--fabric-ids", "999", "--job-ids", "test-job"])
    assert result.exit_code == 1
    assert "[ERROR]: Can't combine filters, Please pass either --fabric_ids or --job_id" in result.output


def test_deploy_with_job_id_filter_and_skip_builds():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH, "--job-ids", "test-job", "--skip-builds"])
    assert result.exit_code == 1
    assert (
        "[ERROR]: Can't skip builds for job_id filter,\nas it only builds depending pipelines ,\nPlease pass "
        "either --skip-builds or --job_id filter" in result.output
    )


def test_deploy_path_pipeline_with_job_id_filter():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH, "--job-ids", "test-job"])

    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert (
        "Found 4 pipelines: customers_orders1243 (python), report_top_customers (python),\njoin_agg_sort (python), "
        "farmers-markets-irs (python)" in result.output
    )
    assert "Deploying jobs only for given Job IDs: ['test-job']" in result.output
    assert "[INFO]: Total Unique pipelines dependencies found: 3" in result.output
    assert "[INFO]: Building given custom pipelines" in result.output
    assert "[INFO]: Generating depending pipelines for all jobs" in result.output
    assert "Building 3 pipelines" in result.output
    assert "Building pipeline pipelines/customers_orders" in result.output
    assert "Building pipeline pipelines/report_top_customers" in result.output
    assert "Building pipeline pipelines/join_agg_sort" in result.output
    assert "Deploying 1 jobs" in result.output
    assert "[START]:  Deploying job jobs/test-job" in result.output
    assert "Deploying job jobs/job-another" not in result.output


def test_deploy_path_pipeline_with_multiple_job_id_filter():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH, "--job-ids", "test-job,job-another"])

    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert (
        "Found 4 pipelines: customers_orders1243 (python), report_top_customers (python),\njoin_agg_sort (python), "
        "farmers-markets-irs (python)" in result.output
    )
    assert "Deploying jobs only for given Job IDs: ['test-job', 'job-another']" in result.output
    assert "[INFO]: Total Unique pipelines dependencies found: 4" in result.output
    assert "[INFO]: Building given custom pipelines" in result.output
    assert "Building 4 pipelines" in result.output
    assert "Building pipeline pipelines/customers_orders" in result.output
    assert "Building pipeline pipelines/report_top_customers" in result.output
    assert "Building pipeline pipelines/join_agg_sort" in result.output
    assert "Building pipeline pipelines/farmers-markets-irs" in result.output
    assert "Deploying 2 jobs" in result.output
    assert "[START]:  Deploying job jobs/test-job" in result.output
    assert "[START]:  Deploying job jobs/job-another" in result.output


def test_deploy_path_pipeline_with_one_invalid_job_id_filter():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH, "--job-ids", "invalid1,test-job"])

    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert (
        "Found 4 pipelines: customers_orders1243 (python), report_top_customers (python),\njoin_agg_sort (python), "
        "farmers-markets-irs (python)" in result.output
    )
    assert "Deploying jobs only for given Job IDs: ['invalid1', 'test-job']" in result.output
    assert "[INFO]: Total Unique pipelines dependencies found: 3" in result.output
    assert "[INFO]: Building given custom pipelines" in result.output
    assert "Building 3 pipelines" in result.output
    assert "Building pipeline pipelines/customers_orders" in result.output
    assert "Building pipeline pipelines/report_top_customers" in result.output
    assert "Building pipeline pipelines/join_agg_sort" in result.output
    assert "Deploying 1 jobs" in result.output
    assert "[START]:  Deploying job jobs/test-job" in result.output


def test_deploy_path_pipeline_with_all_invalid_job_ids_filter():
    runner = CliRunner()
    result = runner.invoke(deploy, ["--path", PROJECT_PATH, "--job-ids", "invalid1,invalid2"])

    assert result.exit_code == 1
    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert (
        "Found 4 pipelines: customers_orders1243 (python), report_top_customers (python),\njoin_agg_sort (python), "
        "farmers-markets-irs (python)" in result.output
    )
    assert "Deploying jobs only for given Job IDs: ['invalid1', 'invalid2']" in result.output
    assert (
        "[ERROR]: No Job IDs matches with passed --job_id filter ['invalid1', 'invalid2']\nAvailable Job IDs are: "
        "dict_keys(['jobs/test-job', 'jobs/job-another']" in result.output
    )



import os
import re
import subprocess
import threading
from typing import List

from ..entities.project import Project
from ..utils.project_config import ProjectConfig
from ..utils.project_models import StepMetadata, StepType, Operation, Status
from ..utility import custom_print as log, Either

GEMS = "Gems"


class GemsDeployment:
    def __init__(self, project: Project, project_config: ProjectConfig):
        self.project = project
        self.project_config = project_config

    def _does_gems_exist(self):
        return self.project.non_empty_gems_directory()

    def summary(self):
        if self._does_gems_exist():
            return ["Gems package will be build and uploaded to nexus"]
        else:
            return []

    def headers(self) -> List[StepMetadata]:
        if self._does_gems_exist():
            return [StepMetadata(GEMS, "Gems will be built and uploaded",
                                 Operation.Build, StepType.Pipeline)]
        else:
            return []

    def deploy(self):
        if self._does_gems_exist():

            log(step_status=Status.RUNNING, step_id=GEMS)
            gem_path = os.path.join(self.project.project_path, "gems")
            package_builder = PackageBuilder(gem_path, self.project.project_language)

            try:
                return_code = package_builder.build()

                if return_code == 0:
                    log(step_status=Status.SUCCEEDED, step_id=GEMS)
                    return [Either(right=True)]
                else:
                    log(step_status=Status.FAILED, step_id=GEMS)
                    return [Either(left=Exception(f"Failed to build the gems package with exit code {return_code}."))]

            except Exception as e:
                log(message="Failed to build the pipeline package.", exception=e, step_id=GEMS)
                log(step_id=GEMS, step_status=Status.FAILED)
                return [Either(left=e)]
        else:
            return []


class PackageBuilder:

    def __init__(self, path: str, language: str):

        self.path = path
        self.language = language

    def build(self):
        if self.language == "scala":
            return self.mvn_build()
        else:
            return self.wheel_build()

    def mvn_build(self):
        command = ["mvn", "deploy", "-DskipTests"]

        log(f"Running mvn command {command}", step_id=GEMS)

        return self._build(command)

    def wheel_build(self):
        command = ["python3", "setup.py", "bdist_wheel", "upload", "-r", "local"]

        log(f"Running python command {command}", step_id=GEMS)

        return self._build(command)

    # maybe we can try another iteration with yield ?
    def _build(self, command: list):
        env = dict(os.environ)

        # Set the MAVEN_OPTS variable
        env["MAVEN_OPTS"] = "-Xmx1024m -XX:MaxPermSize=512m -Xss32m"

        process = subprocess.Popen(command, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env,
                                   cwd=self.path)

        def log_output(pipe, log_function):
            while True:
                # Read line from stdout or stderr, break if EOF
                output = pipe.readline()
                if process.poll() is not None and not output:
                    break
                # Decode line and print it
                response = output.decode().strip()

                # stripping unnecessary logs
                if not re.search(r'Progress \(\d+\):', response):
                    log_function(response)

        # Create threads to read and log stdout and stderr simultaneously
        stdout_thread = threading.Thread(target=log_output,
                                         args=(process.stdout, lambda msg: log(msg, step_id=GEMS)))
        stderr_thread = threading.Thread(target=log_output,
                                         args=(process.stderr, lambda msg: log(msg, step_id=GEMS)))

        # Start threads
        stdout_thread.start()
        stderr_thread.start()

        # Wait for both threads to finish
        stdout_thread.join()
        stderr_thread.join()

        # Get the exit code
        return_code = process.wait()

        if return_code == 0:
            log("Build was successful.", step_id=GEMS)
        else:
            log(f"Build failed with exit code {return_code}", step_id=GEMS)

        return return_code



import os
from enum import Enum
from typing import Optional, Any

from .utils.project_models import LogEvent, StepMetadata, Status, LogLevel


# class mimicking Either behaviour to capture both success and failure
class Either:

    def __init__(self, left=None, right=None):
        self.left = left
        self.right = right

    @property
    def is_left(self):
        return self.left is not None

    @property
    def is_right(self) -> bool:
        return self.right is not None


def custom_print(message: Optional[Any] = None, exception=None,
                 step_id=None,
                 step_metadata: Optional[StepMetadata] = None,
                 step_status: Optional[Status] = None, level: LogLevel = LogLevel.INFO, indent: int = 0) -> None:
    if is_online_mode():
        # Custom print: Print all variables.
        if step_metadata is not None:
            log_event = LogEvent.from_step_metadata(step_metadata)
        elif step_status is not None:
            log_event = LogEvent.from_status(step_status, step_id)
        else:
            log_event = LogEvent.from_log(step_id, message, exception)

        print(log_event.to_json(), flush=True)
    else:
        prefix_space = ''
        if indent != 0:
            prefix_space = ' ' * indent
        # Regular print: Skip stepName.
        if exception is not None:
            print(f'{prefix_space}{message}', exception)
        else:
            if message is not None:
                print(f'{prefix_space}{message}')


def is_online_mode() -> bool:
    return os.environ.get("PRINT_MODE", "REGULAR") == "CUSTOM"


# If the item is a dictionary


def remove_null_items_recursively(item):
    # If the item is a dictionary
    if isinstance(item, dict):
        return {
            k: remove_null_items_recursively(v)
            for k, v in item.items() if v is not None
        }

    # If the item is a list
    elif isinstance(item, list):
        return [remove_null_items_recursively(v) for v in item if v is not None]

    # If the item is an enum
    elif isinstance(item, Enum):
        return item.value

    # If the item is neither a dictionary, a list, nor an enum
    else:
        return item



"""
DATABRICKS_HOST, DATABRICKS_TOKEN
"""
from typing import Optional

import click
import pkg_resources
from rich import print

from .pbt_cli import PBTCli
from .prophecy_build_tool import ProphecyBuildTool
from .utility import is_online_mode


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--path",
    help="Path to the directory containing the pbt_project.yml file",
    required=True,
)
@click.option(
    "--pipelines",
    help="Pipeline names(comma separated) which can be used to filter pipelines to be build",
    default="",
)
@click.option(
    "--ignore-build-errors",
    help="Flag to ignore any build errors in pipelines and return success (EXIT_CODE = 0), please refer logs for any "
         "errors.",
    default=False,
    is_flag=True,
    required=False,
)
@click.option(
    "--ignore-parse-errors",
    help="Flag to ignore any parsing errors in pipelines and return success (EXIT_CODE = 0), please refer logs for "
         "any errors.",
    default=False,
    is_flag=True,
    required=False,
)
def build(path, pipelines, ignore_build_errors, ignore_parse_errors):
    pbt = ProphecyBuildTool(path, ignore_parse_errors=ignore_parse_errors)
    pbt.build(pipelines, exit_on_build_failure=(not ignore_build_errors))


@cli.command()
@click.option(
    "--path",
    help="Path to the directory containing the pbt_project.yml file",
    required=True,
)
@click.option(
    "--pipelines",
    help="Pipeline names(comma separated) which can be used to filter pipelines to be build",
    default="",
)
@click.option(
    "--ignore-build-errors",
    help="Flag to ignore any build errors in pipelines and return success (EXIT_CODE = 0), please refer logs for any "
         "errors.",
    default=False,
    is_flag=True,
    required=False,
)
@click.option(
    "--ignore-parse-errors",
    help="Flag to ignore any parsing errors in pipelines and return success (EXIT_CODE = 0), please refer logs for "
         "any errors.",
    default=False,
    is_flag=True,
    required=False,
)
def build_v2(path, pipelines, ignore_build_errors, ignore_parse_errors):
    pbt = PBTCli.from_conf_folder(path)
    pbt.build(pipelines, ignore_build_errors, ignore_parse_errors)


@cli.command()
@click.option(
    "--path",
    help="Path to the directory containing the pbt_project.yml file",
    required=True,
)
@click.option(
    "--treat-warnings-as-errors",
    help="Specifies whether to treat warnings as errors.",
    is_flag=True,
    required=False,
    default=False,
)
def validate(path, treat_warnings_as_errors):
    pbt = ProphecyBuildTool(path)
    pbt.validate(treat_warnings_as_errors)


@cli.command()
@click.option(
    "--path",
    help="Path to the directory containing the pbt_project.yml file",
    required=True,
)
@click.option(
    "--treat-warnings-as-errors",
    help="Specifies whether to treat warnings as errors.",
    is_flag=True,
    required=False,
    default=False,
)
def validate_v2(path, treat_warnings_as_errors):
    pbt = PBTCli.from_conf_folder(path)
    pbt.validate(treat_warnings_as_errors)


@cli.command()
@click.option(
    "--path",
    help="Path to the directory containing the pbt_project.yml file",
    required=True,
)
@click.option("--dependent-projects-path", help="Dependent projects path", default="")
@click.option(
    "--release-version",
    help="Release version to be used during deployments",
    default="",
)
@click.option(
    "--project-id",
    help="Project Id placeholder to be used during deployments",
    default="",
)
@click.option(
    "--prophecy-url",
    help="Prophecy URL placeholder to be used during deployments",
    default="",
)
@click.option(
    "--fabric-ids",
    help="Fabric IDs(comma separated) which can be used to filter jobs for deployments",
    default="",
)
@click.option(
    "--job-ids",
    help="Job IDs(comma separated) which can be used to filter jobs for deployment",
    default="",
)
@click.option("--skip-builds", is_flag=True, default=False, help="Flag to skip building Pipelines")
def deploy(
        path,
        dependent_projects_path,
        release_version,
        project_id,
        prophecy_url,
        fabric_ids,
        job_ids,
        skip_builds,
):
    pbt = ProphecyBuildTool(path, dependent_projects_path, release_version, project_id, prophecy_url)
    pbt.deploy(fabric_ids=fabric_ids, skip_builds=skip_builds, job_ids=job_ids)


@cli.command()
@click.option(
    "--path",
    help="Path to the directory containing the pbt_project.yml file",
    required=True,
)
@click.option(
    "--project-id",
    help="Path to the directory containing the pbt_project.yml file",
    default="1",
    required=False,
)
@click.option(
    "--conf-dir",
    help="Path to the configuration file folders",
    default="",
    required=False,
)
@click.option(
    "--release-tag",
    help="Release tag",
    default="1.0",
    required=False,
)
@click.option(
    "--release-version",
    help="Release version",
    default="1.0",
    required=False,
)
@click.option(
    "--fabric-ids",
    help="Fabric IDs(comma separated) which can be used to filter jobs for deployments",
    default="",
)
@click.option(
    "--job-ids",
    help="Job IDs(comma separated) which can be used to filter jobs for deployment",
    default="",
)
@click.option(
    "--skip-builds",
    default=False,
    is_flag=True,
    help="Flag to skip building Pipelines",
    required=False
)
@click.option(
    "--dependent-projects-path",
    default="",
    help="Dependent projects path",
    required=False
)
@click.option(
    "--migrate",
    default=False,
    is_flag=True,
    help="Migrate v1 to v2 based project",
    required=False
)
def deploy_v2(path: str,
              project_id: str,
              conf_dir: Optional[str],
              release_tag: Optional[str],
              release_version: str,
              fabric_ids: str,
              job_ids: str,
              skip_builds: bool, dependent_projects_path: str, migrate: bool):
    pbt = PBTCli.from_conf_folder(path, project_id, conf_dir, release_tag, release_version, fabric_ids, job_ids,
                                  skip_builds, dependent_projects_path, migrate)
    if is_online_mode():
        pbt.headers()
    pbt.deploy([])


@cli.command()
@click.option(
    "--path",
    help="Path to the directory containing the pbt_project.yml file",
    required=True,
)
@click.option(
    "--driver-library-path",
    help="Jar path of prophecy-python-libs and other required dependencies",
    required=False,
)
def test_v2(path, driver_library_path):
    pbt = PBTCli.from_conf_folder(path)
    pbt.test(driver_library_path)


@cli.command()
@click.option(
    "--path",
    help="Path to the directory containing the pbt_project.yml file",
    required=True,
)
@click.option(
    "--driver-library-path",
    help="Jar path of prophecy-python-libs and other required dependencies",
    required=False,
)
@click.option(
    "--pipelines",
    help="Pipeline names(comma separated) which can be used to filter pipelines to be tested",
    default="",
)
def test(path, driver_library_path, pipelines):
    pbt = ProphecyBuildTool(path)
    pbt.test(driver_library_path, pipelines)


if __name__ == "pbt":
    print(
        f"[bold purple]Prophecy-build-tool[/bold purple] [bold black]"
        f"v{pkg_resources.require('prophecy-build-tool')[0].version}[/bold black]\n"
    )
    cli()
